"""
Access data provided by the `United States Army Corps of Engineers`_.

Access data from the USACE `Tulsa District Water Control`_ web site.

.. _United States Army Corps of Engineers: http://www.usace.army.mil/
.. _Tulsa District Water Control: http://www.swt-wc.usace.army.mil/

"""
import datetime

import pandas as pd

from tsgettoolbox.ulmo.usace.swtwc.core import get_station_data

# def get_station_data(station_code, date=None, as_dataframe=False):


def ulmo_df(station_code, date=None):
    if date is None:
        date = datetime.datetime.now()
    else:
        date = pd.to_datetime(date)
    alldict = get_station_data(station_code, date=date, as_dataframe=True)
    df = alldict["values"]
    df.columns = [
        "{0}:{1}".format(i, alldict["variables"][i]["unit"]) for i in df.columns
    ]
    df.columns = [i.replace("  ", "_").replace(" ", "_") for i in df.columns]
    df = df.tz_localize(alldict["timezone"])
    return df


if __name__ == "__main__":
    #    import time
    #
    #    r = ulmo_df('blah',
    #                'upperbasin')
    #
    #    print('UB EVERYTHING')
    #    print(r)
    #
    #    time.sleep(20)

    r = ulmo_df("PTTK1", date="2017-11-30")

    print("PTTK1 EVERYTHING")
    print(r)



"""
    `USGS National Water Information System`_ web services


    .. _USGS National Water Information System: http://waterdata.usgs.gov/nwis
"""
from __future__ import absolute_import

from . import core

from .core import get_sites, get_site_data

from tsgettoolbox.ulmo import util

try:
    from . import hdf5

except ImportError:
    hdf5 = util.module_with_dependency_errors(
        [
            "get_site",
            "get_sites",
            "get_site_data",
            "update_site_list",
            "update_site_data",
        ]
    )

pytables = util.module_with_deprecation_warnings(
    [
        hdf5.get_site,
        hdf5.get_sites,
        hdf5.get_site_data,
        hdf5.update_site_list,
        hdf5.update_site_data,
    ],
    "the nwis.pytables module has moved to nwis.hdf5 - nwis.pytables "
    "is deprecated and will be removed in a future ulmo release.",
)



from __future__ import absolute_import, division, print_function

import pytest

pytest.importorskip("sqlalchemy")

import os
from decimal import Decimal
from functools import partial
from textwrap import dedent

import datashape
from datashape import (
    date_,
    datetime_,
    discover,
    dshape,
    int_,
    int64,
    float32,
    float64,
    string,
    var,
    Option,
    R,
)
from datashape.util.testing import assert_dshape_equal
import numpy as np
import pandas as pd
import sqlalchemy as sa

from tsgettoolbox.odo import convert, append, resource, into, odo, chunks
from tsgettoolbox.odo.backends.sql import (
    dshape_to_table,
    create_from_datashape,
    dshape_to_alchemy,
    discover_sqlalchemy_selectable,
)
from tsgettoolbox.odo.utils import tmpfile, raises

from six import string_types


def test_resource():
    sql = resource("sqlite:///:memory:::mytable", dshape="var * {x: int, y: int}")
    assert isinstance(sql, sa.Table)
    assert sql.name == "mytable"
    assert isinstance(sql.bind, sa.engine.base.Engine)
    assert set(c.name for c in sql.c) == set(["x", "y"])


def test_append_and_convert_round_trip():
    engine = sa.create_engine("sqlite:///:memory:")
    metadata = sa.MetaData(engine)
    t = sa.Table(
        "bank",
        metadata,
        sa.Column("name", sa.String, primary_key=True),
        sa.Column("balance", sa.Integer),
    )
    t.create()

    data = [("Alice", 1), ("Bob", 2)]
    append(t, data)

    assert convert(list, t) == data


def test_plus_must_have_text():
    with pytest.raises(NotImplementedError):
        resource("redshift+://user:pass@host:1234/db")


def test_resource_on_file():
    with tmpfile(".db") as fn:
        uri = "sqlite:///" + fn
        sql = resource(uri, "foo", dshape="var * {x: int, y: int}")
        assert isinstance(sql, sa.Table)

    with tmpfile(".db") as fn:
        uri = "sqlite:///" + fn
        sql = resource(uri + "::" + "foo", dshape="var * {x: int, y: int}")
        assert isinstance(sql, sa.Table)


def test_resource_to_engine():
    with tmpfile(".db") as fn:
        uri = "sqlite:///" + fn
        r = resource(uri)
        assert isinstance(r, sa.engine.Engine)
        assert r.dialect.name == "sqlite"


def test_resource_to_engine_to_create_tables():
    with tmpfile(".db") as fn:
        uri = "sqlite:///" + fn
        ds = datashape.dshape("{mytable: var * {name: string, amt: int}}")
        r = resource(uri, dshape=ds)
        assert isinstance(r, sa.engine.Engine)
        assert r.dialect.name == "sqlite"

        assert discover(r) == ds


def test_discover():
    assert discover(sa.String()) == datashape.string
    metadata = sa.MetaData()
    s = sa.Table(
        "accounts",
        metadata,
        sa.Column("name", sa.String),
        sa.Column("amount", sa.Integer),
        sa.Column("timestamp", sa.DateTime, primary_key=True),
    )
    ds = dshape("var * {name: ?string, amount: ?int32, timestamp: datetime}")
    assert_dshape_equal(discover(s), ds)
    for name in ds.measure.names:
        assert isinstance(name, string_types)


def test_discover_numeric_column():
    assert discover(sa.String()) == datashape.string
    metadata = sa.MetaData()
    s = sa.Table("name", metadata, sa.Column("name", sa.types.NUMERIC))

    assert discover(s)


def test_discover_null_columns():
    assert dshape(discover(sa.Column("name", sa.String, nullable=True))) == dshape(
        "{name: ?string}"
    )
    assert dshape(discover(sa.Column("name", sa.String, nullable=False))) == dshape(
        "{name: string}"
    )


def test_discover_selectable():
    t = resource("sqlite:///:memory:::mytable", dshape="var * {x: int, y: int}")
    q = sa.select([t.c.x]).limit(5)
    assert discover(q) == dshape("var * {x: int}")


def test_discover_fixed_length_string():
    t = resource("sqlite:///:memory:::mytable", dshape="var * {x: string[30]}")
    assert discover(t) == dshape("var * {x: string[30]}")


def single_table_engine():
    engine = sa.create_engine("sqlite:///:memory:")
    metadata = sa.MetaData(engine)
    t = sa.Table(
        "accounts",
        metadata,
        sa.Column("name", sa.String),
        sa.Column("amount", sa.Integer),
    )
    t.create()
    return engine, t


def test_select_to_iterator():
    engine, t = single_table_engine()
    append(t, [("Alice", 100), ("Bob", 200)])

    sel = sa.select([t.c.amount + 1])

    assert convert(list, sel) == [(101,), (201,)]
    assert convert(list, sel, dshape=dshape("var * int")) == [101, 201]

    sel2 = sa.select([sa.sql.func.sum(t.c.amount)])

    assert convert(int, sel2, dshape=dshape("int")) == 300

    sel3 = sa.select([t])

    result = convert(list, sel3, dshape=discover(t))
    assert type(result[0]) is tuple
    for res in result:
        assert isinstance(res[0], string_types)


def test_discovery_engine():
    engine, t = single_table_engine()

    assert discover(engine, "accounts") == discover(t)

    assert str(discover(engine)) == str(discover({"accounts": t}))


def test_discovery_metadata():
    engine, t = single_table_engine()
    metadata = t.metadata
    assert str(discover(metadata)) == str(discover({"accounts": t}))


def test_discover_views():
    engine, t = single_table_engine()
    metadata = t.metadata
    with engine.connect() as conn:
        conn.execute(
            """CREATE VIEW myview AS
                        SELECT name, amount
                        FROM accounts
                        WHERE amount > 0"""
        )

    assert str(discover(metadata)) == str(discover({"accounts": t, "myview": t}))


def test_extend_empty():
    engine, t = single_table_engine()

    assert not convert(list, t)
    append(t, [])
    assert not convert(list, t)


def test_dshape_to_alchemy():
    assert dshape_to_alchemy("string") == sa.Text
    assert isinstance(dshape_to_alchemy("string[40]"), sa.String)
    assert not isinstance(dshape_to_alchemy('string["ascii"]'), sa.Unicode)
    assert isinstance(dshape_to_alchemy('string[40, "U8"]'), sa.Unicode)
    assert dshape_to_alchemy("string[40]").length == 40

    assert dshape_to_alchemy("float32") == sa.REAL
    assert dshape_to_alchemy("float64") == sa.FLOAT


def test_dshape_to_table():
    t = dshape_to_table("bank", "{name: string, amount: int}")
    assert isinstance(t, sa.Table)
    assert t.name == "bank"
    assert [c.name for c in t.c] == ["name", "amount"]


td_freqs = list(
    zip(
        ["D", "h", "m", "s", "ms", "us", "ns"],
        [0, 0, 0, 0, 3, 6, 9],
        [9, 0, 0, 0, 0, 0, 0],
    )
)


@pytest.mark.parametrize(["freq", "secp", "dayp"], td_freqs)
def test_dshape_to_table_with_timedelta(freq, secp, dayp):
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table("td_bank", ds)
    assert isinstance(t, sa.Table)
    assert t.name == "td_bank"
    assert isinstance(t.c.duration.type, sa.types.Interval)
    assert t.c.duration.type.second_precision == secp
    assert t.c.duration.type.day_precision == dayp


@pytest.mark.xfail(raises=NotImplementedError)
def test_dshape_to_table_month():
    ds = '{name: string, amount: int, duration: timedelta[unit="M"]}'
    dshape_to_table("td_bank", ds)


@pytest.mark.xfail(raises=NotImplementedError)
def test_dshape_to_table_year():
    ds = '{name: string, amount: int, duration: timedelta[unit="Y"]}'
    dshape_to_table("td_bank", ds)


@pytest.mark.parametrize("freq", ["D", "s", "ms", "us", "ns"])
def test_timedelta_sql_discovery(freq):
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table("td_bank", ds)
    assert discover(t).measure["duration"] == datashape.TimeDelta(freq)


@pytest.mark.parametrize("freq", ["h", "m"])
def test_timedelta_sql_discovery_hour_minute(freq):
    # these always compare equal to a seconds timedelta, because no data loss
    # will occur with this heuristic. this implies that the sa.Table was
    # constructed with day_precision == 0 and second_precision == 0
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table("td_bank", ds)
    assert discover(t).measure["duration"] == datashape.TimeDelta("s")


prec = {"s": 0, "ms": 3, "us": 6, "ns": 9}


@pytest.mark.parametrize("freq", list(prec.keys()))
def test_discover_postgres_intervals(freq):
    precision = prec.get(freq)
    typ = sa.dialects.postgresql.base.INTERVAL(precision=precision)
    t = sa.Table("t", sa.MetaData(), sa.Column("dur", typ))
    assert discover(t) == dshape('var * {dur: ?timedelta[unit="%s"]}' % freq)


# between postgresql and oracle, only oracle has support for day intervals


@pytest.mark.parametrize("freq", ["D"] + list(prec.keys()))
def test_discover_oracle_intervals(freq):
    typ = sa.dialects.oracle.base.INTERVAL(
        day_precision={"D": 9}.get(freq), second_precision=prec.get(freq, 0)
    )
    t = sa.Table("t", sa.MetaData(), sa.Column("dur", typ))
    assert discover(t) == dshape('var * {dur: ?timedelta[unit="%s"]}' % freq)


@pytest.mark.parametrize(
    "typ,dtype",
    (
        (sa.DATETIME, datetime_),
        (sa.TIMESTAMP, datetime_),
        (sa.FLOAT, float64),
        (sa.DATE, date_),
        (sa.BIGINT, int64),
        (sa.INTEGER, int_),
        (sa.BIGINT, int64),
        (sa.types.NullType, string),
        (sa.REAL, float32),
        (sa.Float, float64),
        (sa.Float(precision=8), float32),
        (sa.Float(precision=24), float32),
        (sa.Float(precision=42), float64),
        (sa.Float(precision=53), float64),
    ),
)
def test_types(typ, dtype):
    expected = var * R["value" : Option(dtype)]
    t = sa.Table("t", sa.MetaData(), sa.Column("value", typ))
    assert_dshape_equal(discover(t), expected)


@pytest.mark.parametrize(
    "typ", (sa.Float(precision=-1), sa.Float(precision=0), sa.Float(precision=54))
)
def test_unsupported_precision(typ):
    t = sa.Table("t", sa.MetaData(), sa.Column("value", typ))
    with pytest.raises(ValueError) as err:
        discover(t)
    assert str(err.value) == "{} is not a supported precision".format(typ.precision)


def test_mssql_types():
    typ = sa.dialects.mssql.BIT()
    t = sa.Table("t", sa.MetaData(), sa.Column("bit", typ))
    assert_dshape_equal(discover(t), dshape("var * {bit: ?bool}"))
    typ = sa.dialects.mssql.DATETIMEOFFSET()
    t = sa.Table("t", sa.MetaData(), sa.Column("dt", typ))
    assert_dshape_equal(discover(t), dshape("var * {dt: ?string}"))
    typ = sa.dialects.mssql.MONEY()
    t = sa.Table("t", sa.MetaData(), sa.Column("money", typ))
    assert_dshape_equal(discover(t), dshape("var * {money: ?float64}"))
    typ = sa.dialects.mssql.SMALLMONEY()
    t = sa.Table("t", sa.MetaData(), sa.Column("money", typ))
    assert_dshape_equal(discover(t), dshape("var * {money: ?float32}"))
    typ = sa.dialects.mssql.UNIQUEIDENTIFIER()
    t = sa.Table("t", sa.MetaData(), sa.Column("uuid", typ))
    assert_dshape_equal(discover(t), dshape("var * {uuid: ?string}"))


def test_create_from_datashape():
    engine = sa.create_engine("sqlite:///:memory:")
    ds = dshape(
        """{bank: var * {name: string, amount: int},
                    points: var * {x: int, y: int}}"""
    )
    engine = create_from_datashape(engine, ds)

    assert discover(engine) == ds


def test_into_table_iterator():
    engine = sa.create_engine("sqlite:///:memory:")
    metadata = sa.MetaData(engine)
    t = dshape_to_table("points", "{x: int, y: int}", metadata=metadata)
    t.create()

    data = [(1, 1), (2, 4), (3, 9)]
    append(t, data)

    assert convert(list, t) == data
    assert isinstance(convert(list, t)[0], tuple)

    t2 = dshape_to_table("points2", "{x: int, y: int}", metadata=metadata)
    t2.create()
    data2 = [{"x": 1, "y": 1}, {"x": 2, "y": 4}, {"x": 3, "y": 9}]
    append(t2, data2)

    assert convert(list, t2) == data


def test_sql_field_names_disagree_on_order():
    r = resource("sqlite:///:memory:::tb", dshape=dshape("{x: int, y: int}"))
    append(r, [(1, 2), (10, 20)], dshape=dshape("{y: int, x: int}"))
    assert convert(set, r) == set([(2, 1), (20, 10)])


def test_sql_field_names_disagree_on_names():
    r = resource("sqlite:///:memory:::tb", dshape=dshape("{x: int, y: int}"))
    assert raises(
        Exception,
        lambda: append(r, [(1, 2), (10, 20)], dshape=dshape("{x: int, z: int}")),
    )


def test_resource_on_dialects():
    assert resource.dispatch("mysql://foo") is resource.dispatch("mysql+pymysql://foo")
    assert resource.dispatch("never-before-seen-sql://foo") is resource.dispatch(
        "mysql://foo"
    )


@pytest.yield_fixture
def sqlite_file():
    try:
        yield "sqlite:///db.db"
    finally:
        os.remove("db.db")


def test_append_from_select(sqlite_file):
    # we can't test in memory here because that creates two independent
    # databases
    raw = np.array(
        [(200.0, "Glenn"), (314.14, "Hope"), (235.43, "Bob")],
        dtype=[("amount", "float64"), ("name", "U5")],
    )
    raw2 = np.array(
        [(800.0, "Joe"), (914.14, "Alice"), (1235.43, "Ratso")],
        dtype=[("amount", "float64"), ("name", "U5")],
    )
    t = into("%s::t" % sqlite_file, raw)
    s = into("%s::s" % sqlite_file, raw2)
    t = append(t, s.select())
    result = into(list, t)
    expected = np.concatenate((raw, raw2)).tolist()
    assert result == expected


def test_append_from_table():
    # we can't test in memory here because that creates two independent
    # databases
    with tmpfile("db") as fn:
        raw = np.array(
            [(200.0, "Glenn"), (314.14, "Hope"), (235.43, "Bob")],
            dtype=[("amount", "float64"), ("name", "U5")],
        )
        raw2 = np.array(
            [(800.0, "Joe"), (914.14, "Alice"), (1235.43, "Ratso")],
            dtype=[("amount", "float64"), ("name", "U5")],
        )
        t = into("sqlite:///%s::t" % fn, raw)
        s = into("sqlite:///%s::s" % fn, raw2)
        t = append(t, s)
        result = odo(t, list)
        expected = np.concatenate((raw, raw2)).tolist()
        assert result == expected


def test_engine_metadata_caching():
    with tmpfile("db") as fn:
        engine = resource("sqlite:///" + fn)
        a = resource("sqlite:///" + fn + "::a", dshape=dshape("var * {x: int}"))
        b = resource("sqlite:///" + fn + "::b", dshape=dshape("var * {y: int}"))

        assert a.metadata is b.metadata
        assert engine is a.bind is b.bind


def test_copy_one_table_to_a_foreign_engine():
    data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape("var * {x: int, y: int}")
    with tmpfile("db") as fn1:
        with tmpfile("db") as fn2:
            src = into("sqlite:///%s::points" % fn1, data, dshape=ds)
            tgt = into("sqlite:///%s::points" % fn2, sa.select([src]), dshape=ds)

            assert into(set, src) == into(set, tgt)
            assert into(set, data) == into(set, tgt)


def test_select_to_series_retains_name():
    data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape("var * {x: int, y: int}")
    with tmpfile("db") as fn1:
        points = odo(data, "sqlite:///%s::points" % fn1, dshape=ds)
        sel = sa.select([(points.c.x + 1).label("x")])
        series = odo(sel, pd.Series)
    assert series.name == "x"
    assert odo(series, list) == [x + 1 for x, _ in data]


def test_empty_select_to_empty_frame():
    # data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape("var * {x: int, y: int}")
    with tmpfile("db") as fn1:
        points = resource("sqlite:///%s::points" % fn1, dshape=ds)
        sel = sa.select([points])
        df = odo(sel, pd.DataFrame)
    assert df.empty
    assert df.columns.tolist() == ["x", "y"]


def test_discover_foreign_keys():
    with tmpfile("db") as fn:
        products = resource(
            "sqlite:///%s::products" % fn,
            dshape="""
                                var * {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
            primary_key=["product_no"],
        )
        expected = dshape(
            """var * {
                          order_id: int32,
                          product_no: map[int32, {
                            product_no: int32,
                            name: ?string,
                            price: ?float64
                          }],
                          quantity: ?int32
                        }"""
        )
        orders = resource(
            "sqlite:///%s::orders" % fn,
            dshape=expected,
            foreign_keys=dict(product_no=products.c.product_no),
        )
        result = discover(orders)
        assert result == expected


def test_invalid_foreign_keys():
    with tmpfile("db") as fn:
        expected = dshape(
            """var * {
                          order_id: int32,
                          product_no: map[int32, {
                            product_no: int32,
                            name: ?string,
                            price: ?float64
                          }],
                          quantity: ?int32
                        }"""
        )
        with pytest.raises(TypeError):
            resource("sqlite:///%s::orders" % fn, dshape=expected)


def test_foreign_keys_auto_construct():
    with tmpfile("db") as fn:
        products = resource(
            "sqlite:///%s::products" % fn,
            dshape="""
                                var * {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
            primary_key=["product_no"],
        )
        ds = dshape(
            """var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          quantity: ?int32
                        }"""
        )
        orders = resource(
            "sqlite:///%s::orders" % fn,
            dshape=ds,
            foreign_keys=dict(product_no=products.c.product_no),
            primary_key=["order_id"],
        )
        assert discover(orders) == dshape(
            """
            var * {
                order_id: int32,
                product_no: map[int32, {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }],
                quantity: ?int32
            }
        """
        )


def test_foreign_keys_bad_field():
    with tmpfile("db") as fn:
        expected = dshape(
            """var * {
                          order_id: int32,
                          product_no: int64,
                          quantity: ?int32
                        }"""
        )
        with pytest.raises(TypeError):
            resource(
                "sqlite:///%s::orders" % fn,
                dshape=expected,
                foreign_keys=dict(foo="products.product_no"),
            )


@pytest.fixture
def recursive_fkey():
    return sa.Table(
        "employees",
        sa.MetaData(),
        sa.Column("eid", sa.BIGINT, primary_key=True),
        sa.Column("name", sa.TEXT),
        sa.Column("mgr_eid", sa.BIGINT, sa.ForeignKey("employees.eid"), nullable=False),
    )


def test_recursive_foreign_key(recursive_fkey):
    expected = dshape(
        """
        var * {
            eid: int64,
            name: ?string,
            mgr_eid: map[int64, {eid: int64, name: ?string, mgr_eid: int64}]
        }
    """
    )
    assert discover(recursive_fkey) == expected


def test_create_recursive_foreign_key():
    with tmpfile(".db") as fn:
        t = resource(
            "sqlite:///%s::employees" % fn,
            dshape="""
                     var * {
                        eid: int64,
                        name: ?string,
                        mgr_eid: map[int64, T]
                     }""",
            foreign_keys=dict(mgr_eid="employees.eid"),
            primary_key=["eid"],
        )
        result = discover(t)
    expected = dshape(
        """
        var * {
            eid: int64,
            name: ?string,
            mgr_eid: map[int64, {eid: int64, name: ?string, mgr_eid: int64}]
        }
    """
    )
    assert result == expected


def test_compound_primary_key():
    with tmpfile("db") as fn:
        products = resource(
            "sqlite:///%s::products" % fn,
            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
            primary_key=["product_no", "product_sku"],
        )
        assert len(products.primary_key) == 2
        assert products.primary_key.columns["product_no"] is products.c.product_no
        assert products.primary_key.columns["product_sku"] is products.c.product_sku


def test_compound_primary_key_with_fkey():
    with tmpfile("db") as fn:
        products = resource(
            "sqlite:///%s::products" % fn,
            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
            primary_key=["product_no", "product_sku"],
        )
        ds = dshape(
            """var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          product_sku: map[int32, U],
                          quantity: ?int32
                        }"""
        )
        orders = resource(
            "sqlite:///%s::orders" % fn,
            dshape=ds,
            primary_key=["order_id"],
            foreign_keys={
                "product_no": products.c.product_no,
                "product_sku": products.c.product_sku,
            },
        )
        assert discover(orders) == dshape(
            """var * {
                order_id: int32,
                product_no: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                product_sku: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                quantity: ?int32
            }
            """
        )


def test_compound_primary_key_with_single_reference():
    with tmpfile("db") as fn:
        products = resource(
            "sqlite:///%s::products" % fn,
            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
            primary_key=["product_no", "product_sku"],
        )
        # TODO: should this fail everywhere? e.g., this fails in postgres, but
        # not in sqlite because postgres doesn't allow partial foreign keys
        # might be best to let the backend handle this
        ds = dshape(
            """var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          quantity: ?int32
                        }"""
        )
        orders = resource(
            "sqlite:///%s::orders" % fn,
            dshape=ds,
            foreign_keys=dict(product_no=products.c.product_no),
            primary_key=["order_id"],
        )
        assert discover(orders) == dshape(
            """var * {
                order_id: int32,
                product_no: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                quantity: ?int32
            }
            """
        )


def test_foreign_keys_as_compound_primary_key():
    with tmpfile("db") as fn:
        suppliers = resource(
            "sqlite:///%s::suppliers" % fn,
            dshape="var * {id: int64, name: string}",
            primary_key=["id"],
        )
        parts = resource(
            "sqlite:///%s::parts" % fn,
            dshape="var * {id: int64, name: string, region: string}",
            primary_key=["id"],
        )
        suppart = resource(
            "sqlite:///%s::suppart" % fn,
            dshape="var * {supp_id: map[int64, T], part_id: map[int64, U]}",
            foreign_keys={"supp_id": suppliers.c.id, "part_id": parts.c.id},
            primary_key=["supp_id", "part_id"],
        )
        expected = dshape(
            """
            var * {
                supp_id: map[int64, {id: int64, name: string}],
                part_id: map[int64, {id: int64, name: string, region: string}]
            }
        """
        )
        result = discover(suppart)
        assert result == expected


def test_append_chunks():
    tbl = resource("sqlite:///:memory:::test", dshape="var * {a: int, b: int}")
    res = odo(
        chunks(np.ndarray)((np.array([[0, 1], [2, 3]]), np.array([[4, 5], [6, 7]]))),
        tbl,
    )
    assert res is tbl
    assert (
        odo(tbl, np.ndarray)
        == np.array(
            [(0, 1), (2, 3), (4, 5), (6, 7)], dtype=[("a", "<i4"), ("b", "<i4")]
        )
    ).all()


def test_append_array_without_column_names():
    with pytest.raises(TypeError):
        odo(np.zeros((2, 2)), "sqlite:///:memory:::test")


def test_numeric_create():
    tbl = resource(
        "sqlite:///:memory:::test",
        dshape="var * {a: ?decimal[11, 2], b: decimal[10, 6]}",
    )
    assert tbl.c.a.nullable
    assert not tbl.c.b.nullable
    assert isinstance(tbl.c.a.type, sa.NUMERIC)
    assert isinstance(tbl.c.b.type, sa.NUMERIC)


def test_numeric_append():
    tbl = resource(
        "sqlite:///:memory:::test",
        dshape="var * {a: decimal[11, 2], b: ?decimal[10, 6]}",
    )
    data = [(1.0, 2.0), (2.0, 3.0)]
    tbl = odo(data, tbl)
    assert odo(tbl, list) == list(
        map(lambda row: tuple(map(Decimal, row)), tbl.select().execute().fetchall())
    )


def test_discover_float_and_real_core_types():
    assert discover(sa.FLOAT()) == float64
    assert discover(sa.REAL()) == float32


def test_string_dshape_doc_example():
    x = np.zeros((10, 2))
    with tmpfile(".db") as fn:
        t = odo(x, "sqlite:///%s::x" % fn, dshape="var * {a: float64, b: float64}")
        assert all(row == (0, 0) for row in t.select().execute().fetchall())


def test_decimal_conversion():
    data = [(1.0,), (2.0,)]
    with tmpfile(".db") as fn:
        t = odo(data, "sqlite:///%s::x" % fn, dshape="var * {x: decimal[11, 2]}")
        result = odo(sa.select([sa.func.sum(t.c.x)]), Decimal)
    assert result == sum(Decimal(r[0]) for r in data)


def test_append_empty_iterator_returns_table():
    with tmpfile(".db") as fn:
        t = resource("sqlite:///%s::x" % fn, dshape="var * {a: int32}")
        assert odo(iter([]), t) is t


def test_pass_non_hashable_arg_to_create_engine():
    with tmpfile(".db") as fn:
        r = partial(
            resource, "sqlite:///%s::t" % fn, connect_args={}, dshape="var * {a: int32}"
        )
        assert r() is r()

    s = partial(
        resource, "sqlite:///:memory:::t", connect_args={}, dshape="var * {a: int32}"
    )
    assert s() is not s()


def test_nullable_foreign_key():
    """Test for issue #554"""
    engine = sa.create_engine("sqlite://")
    metadata = sa.MetaData(bind=engine)

    T1 = sa.Table(
        "NullableForeignKeyDemo",
        metadata,
        sa.Column("pkid", sa.Integer, primary_key=True),
        sa.Column(
            "label_id",
            sa.Integer,
            sa.ForeignKey("ForeignKeyLabels.pkid"),
            nullable=True,
        ),
    )

    T2 = sa.Table(
        "ForeignKeyLabels",
        metadata,
        sa.Column("pkid", sa.Integer, primary_key=True),
        sa.Column("label", sa.String),
    )

    metadata.create_all()

    x = np.arange(10)
    records1 = [
        {"pkid": idx, "label_id": int(value)} for idx, value in enumerate(x[::-1])
    ]
    records1[-1]["label_id"] = None  # foreign-key is nullable!

    records2 = [{"pkid": int(pkid), "label": chr(pkid + 65)} for pkid in x]
    with engine.connect() as conn:
        conn.execute(T1.insert(), records1)
        conn.execute(T2.insert(), records2)

    ds = discover_sqlalchemy_selectable(T1)
    # The nullable key should be an Option instance
    assert isinstance(ds.measure["label_id"].key, Option)

    dtype = [("pkid", np.int32), ("label_id", object)]
    expected = np.rec.fromarrays([x, x[::-1]], dtype=dtype)
    expected = pd.DataFrame(expected)
    expected.iloc[-1, -1] = None

    actual = odo(T1, pd.DataFrame)
    assert actual.equals(expected)


def test_transaction():
    with tmpfile(".db") as fn:
        rsc = resource("sqlite:///%s::table" % fn, dshape="var * {a: int}")

    data = [(1,), (2,), (3,)]

    conn_1 = rsc.bind.connect()
    conn_2 = rsc.bind.connect()

    trans_1 = conn_1.begin()
    conn_2.begin()

    odo(data, rsc, bind=conn_1)

    # inside the transaction the write should be there
    assert odo(rsc, list, bind=conn_1) == data

    # outside of a transaction or in a different transaction the write is not
    # there
    assert odo(rsc, list) == odo(rsc, list, bind=conn_2) == []

    trans_1.commit()

    # now the data should appear outside the transaction
    assert odo(rsc, list) == odo(rsc, list, bind=conn_2) == data


def test_integer_detection():
    """Test for PR #596"""
    engine = sa.create_engine("sqlite://")
    metadata = sa.MetaData(bind=engine)

    T = sa.Table(
        "Demo",
        metadata,
        sa.Column("pkid", sa.Integer, primary_key=True),
        sa.Column(
            "value", sa.DECIMAL(precision=1, scale=0, asdecimal=False), nullable=True
        ),
    )
    metadata.create_all()

    values = [1, 0, 1, 0, None, 1, 1, 1, 0, 1]
    pkids = range(len(values))
    dtype = [("pkid", np.int32), ("value", np.float64)]
    expected = np.array(list(zip(pkids, values)), dtype=dtype)
    expected = pd.DataFrame(expected)
    records = expected.to_dict(orient="records")
    with engine.connect() as conn:
        conn.execute(T.insert(), records)

    actual = odo(T, pd.DataFrame)
    assert actual.equals(expected)



from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from datetime import datetime, timedelta
from distutils.version import LooseVersion

from datashape import (
    Categorical,
    DateTime,
    Option,
    Record,
    discover,
    dshape,
    float64,
    int64,
    coretypes as ct,
)
from datashape.util.testing import assert_dshape_equal
from networkx import NetworkXNoPath
import numpy as np
import pandas as pd
import pytest

from tsgettoolbox.odo import odo


requires_datetimetz = pytest.mark.skipif(
    LooseVersion(pd.__version__) < LooseVersion("0.17"),
    reason="Pandas before DatetimeTZ",
)


data = [("Alice", 100), ("Bob", 200)]


def test_discover_dataframe():
    df = pd.DataFrame([("Alice", 100), ("Bob", 200)], columns=["name", "amount"])

    assert discover(df) == dshape("2 * {name: ?string, amount: int64}")


def test_discover_series():
    s = pd.Series([1, 2, 3])

    assert discover(s) == 3 * discover(s[0])


def test_floats_are_not_optional():
    df = pd.DataFrame([("Alice", 100), ("Bob", None)], columns=["name", "amount"])
    ds = discover(df)
    assert_dshape_equal(ds[1].types[1], float64)


def test_datetime_to_timestamp():
    dt = datetime(2014, 1, 1)
    ts = odo(dt, pd.Timestamp)
    assert isinstance(ts, pd.Timestamp)
    assert ts == pd.Timestamp("2014-01-01")


def test_nan_to_nat():
    assert odo(float("nan"), pd.Timestamp) is pd.NaT
    assert odo(float("nan"), pd.Timedelta) is pd.NaT
    assert odo(np.nan, pd.Timestamp) is pd.NaT
    assert odo(np.nan, pd.Timedelta) is pd.NaT

    with pytest.raises(NetworkXNoPath):
        # Check that only nan can be converted.
        odo(0.5, pd.Timestamp)

    with pytest.raises(NetworkXNoPath):
        # Check that only nan can be converted.
        odo(0.5, pd.Timedelta)


def test_none_to_nat():
    assert odo(None, pd.Timestamp) is pd.NaT
    assert odo(None, pd.Timedelta) is pd.NaT


def test_nat_to_nat():
    assert odo(pd.NaT, pd.Timestamp) is pd.NaT
    assert odo(pd.NaT, pd.Timedelta) is pd.NaT


def test_timedelta_to_pandas():
    assert odo(timedelta(days=1), pd.Timedelta) == pd.Timedelta(days=1)
    assert odo(timedelta(hours=1), pd.Timedelta) == pd.Timedelta(hours=1)
    assert odo(timedelta(seconds=1), pd.Timedelta) == pd.Timedelta(seconds=1)


def test_categorical_pandas():
    df = pd.DataFrame(
        {"x": list("a" * 5 + "b" * 5 + "c" * 5), "y": np.arange(15, dtype=np.int64)},
        columns=["x", "y"],
    )
    df.x = df.x.astype("category")
    assert_dshape_equal(
        discover(df), 15 * Record([("x", Categorical(["a", "b", "c"])), ("y", int64)])
    )
    assert_dshape_equal(discover(df.x), 15 * Categorical(["a", "b", "c"]))


@requires_datetimetz
def test_datetimetz_pandas():
    df = pd.DataFrame(
        OrderedDict(
            [
                ("naive", pd.date_range("2014", periods=5)),
                ("Europe/Moscow", pd.date_range("2014", periods=5, tz="Europe/Moscow")),
                ("UTC", pd.date_range("2014", periods=5, tz="UTC")),
                ("US/Eastern", pd.date_range("2014", periods=5, tz="US/Eastern")),
            ]
        )
    )

    assert_dshape_equal(
        discover(df),
        5
        * Record[
            "naive" : Option(DateTime(tz=None)),
            "Europe/Moscow" : Option(DateTime(tz="Europe/Moscow")),
            "UTC" : Option(DateTime(tz="UTC")),
            "US/Eastern" : Option(DateTime(tz="US/Eastern")),
        ],
    )

    assert_dshape_equal(discover(df.naive), 5 * Option(DateTime(tz=None)))
    for tz in ("Europe/Moscow", "UTC", "US/Eastern"):
        assert_dshape_equal(discover(df[tz]), 5 * Option(DateTime(tz=tz)))


@pytest.mark.parametrize("dtype", ("int64", "float64"))
def test_numeric_index(dtype):
    ix = pd.Index([1, 2, 3], dtype=dtype)
    actual = discover(ix)
    expected = 3 * getattr(ct, dtype)

    assert_dshape_equal(actual, expected)


# @pytest.mark.parametrize(
#    'tz', (
#        None,
#        requires_datetimetz('US/Eastern'),
#        requires_datetimetz('UTC'),
#        requires_datetimetz('Europe/Moscow'),
#    ),
# )
# def test_datetime_index(tz):
#    ix = pd.DatetimeIndex(['2014-01-01', '2014-01-02', '2014-01-03'], tz=tz)
#    actual = discover(ix)
#    expected = 3 * Option(DateTime(tz=tz))
#
#    assert_dshape_equal(actual, expected)



from __future__ import absolute_import, division, print_function

from tsgettoolbox.odo.regex import RegexDispatcher

import re


def test_regex_dispatcher():
    foo = RegexDispatcher("foo")

    @foo.register(r"\d*", priority=9)
    def a(s):
        """ A's docstring """
        return int(s)

    @foo.register(re.compile(r"\D*"))
    def b(s):
        """ B's docstring """
        return s

    @foo.register(r"0\d*", priority=11)
    def c(s):
        """ C's docstring """
        return s

    assert set(foo.funcs.values()) == set([a, b, c])
    assert foo.dispatch("123") == a
    assert foo.dispatch("hello") == b
    assert foo.dispatch("0123") == c

    assert foo("123") == 123
    assert foo("hello") == "hello"
    assert foo("0123") == "0123"

    assert a.__doc__ in foo.__doc__



from __future__ import absolute_import, division, print_function

import datetime
import os
import gzip
import os
import json

from contextlib import contextmanager

import numpy as np
from tsgettoolbox.odo.backends.json import json_dumps
from tsgettoolbox.odo.utils import tmpfile, ignoring
from tsgettoolbox.odo import (
    odo,
    discover,
    JSONLines,
    resource,
    JSON,
    convert,
    append,
    drop,
)
from tsgettoolbox.odo.temp import Temp, _Temp

from datashape import dshape


@contextmanager
def json_file(data):
    with tmpfile(".json") as fn:
        with open(fn, "w") as f:
            json.dump(data, f, default=json_dumps)

        yield fn


@contextmanager
def jsonlines_file(data):
    with tmpfile(".json") as fn:
        with open(fn, "w") as f:
            for item in data:
                json.dump(item, f, default=json_dumps)
                f.write("\n")

        yield fn


dat = [{"name": "Alice", "amount": 100}, {"name": "Bob", "amount": 200}]


def test_discover_json():
    with json_file(dat) as fn:
        j = JSON(fn)
        assert discover(j) == discover(dat)


def test_discover_jsonlines():
    with jsonlines_file(dat) as fn:
        j = JSONLines(fn)
        assert discover(j) == discover(dat)


def test_discover_json_only_includes_datetimes_not_dates():
    data = [
        {"name": "Alice", "dt": datetime.date(2002, 2, 2)},
        {"name": "Bob", "dt": datetime.date(2000, 1, 1)},
    ]
    with json_file(data) as fn:
        j = JSON(fn)
        assert discover(j) == dshape("2 * {dt: datetime, name: string }")


def test_resource():
    with tmpfile("json") as fn:
        assert isinstance(resource("jsonlines://" + fn), JSONLines)
        assert isinstance(resource("json://" + fn), JSON)

        assert isinstance(
            resource(fn, expected_dshape=dshape("var * {a: int}")), JSONLines
        )


def test_resource_guessing():
    with json_file(dat) as fn:
        assert isinstance(resource(fn), JSON)

    with jsonlines_file(dat) as fn:
        assert isinstance(resource(fn), JSONLines)


def test_append_jsonlines():
    with tmpfile("json") as fn:
        j = JSONLines(fn)
        append(j, dat)
        with open(j.path) as f:
            lines = f.readlines()
        assert len(lines) == 2
        assert "Alice" in lines[0]
        assert "Bob" in lines[1]


def test_append_json():
    with tmpfile("json") as fn:
        j = JSON(fn)
        append(j, dat)
        with open(j.path) as f:
            lines = f.readlines()
        assert len(lines) == 1
        assert "Alice" in lines[0]
        assert "Bob" in lines[0]


def test_convert_json_list():
    with json_file(dat) as fn:
        j = JSON(fn)
        assert convert(list, j) == dat


def test_convert_jsonlines():
    with jsonlines_file(dat) as fn:
        j = JSONLines(fn)
        assert convert(list, j) == dat


def test_tuples_to_json():
    ds = dshape("var * {a: int, b: int}")
    with tmpfile("json") as fn:
        j = JSON(fn)

        append(j, [(1, 2), (10, 20)], dshape=ds)
        with open(fn) as f:
            assert '"a": 1' in f.read()

    with tmpfile("json") as fn:
        j = JSONLines(fn)

        append(j, [(1, 2), (10, 20)], dshape=ds)
        with open(fn) as f:
            assert '"a": 1' in f.read()


def test_datetimes():
    from tsgettoolbox.odo import into
    import numpy as np

    data = [
        {"a": 1, "dt": datetime.datetime(2001, 1, 1)},
        {"a": 2, "dt": datetime.datetime(2002, 2, 2)},
    ]
    with tmpfile("json") as fn:
        j = JSONLines(fn)
        append(j, data)

        assert str(into(np.ndarray, j)) == str(into(np.ndarray, data))


def test_json_encoder():
    result = json.dumps(
        [1, datetime.datetime(2000, 1, 1, 12, 30, 0)], default=json_dumps
    )
    assert result == '[1, "2000-01-01T12:30:00Z"]'
    assert json.loads(result) == [1, "2000-01-01T12:30:00Z"]


def test_empty_line():
    text = '{"a": 1}\n{"a": 2}\n\n'  # extra endline
    with tmpfile(".json") as fn:
        with open(fn, "w") as f:
            f.write(text)
        j = JSONLines(fn)
        assert len(convert(list, j)) == 2


def test_multiple_jsonlines():
    a, b = "_test_a1.json", "_test_a2.json"
    try:
        with ignoring(OSError):
            os.remove(a)
        with ignoring(OSError):
            os.remove(b)
        with open(a, "w") as f:
            json.dump(dat, f)
        with open(b"_test_a2.json", "w") as f:
            json.dump(dat, f)
        r = resource("_test_a*.json")
        result = convert(list, r)
        assert len(result) == len(dat) * 2
    finally:
        with ignoring(OSError):
            os.remove(a)
        with ignoring(OSError):
            os.remove(b)


def test_read_gzip_lines():
    with tmpfile("json.gz") as fn:
        f = gzip.open(fn, "wb")
        for item in dat:
            s = json.dumps(item).encode("utf-8")
            f.write(s)
            f.write(b"\n")
        f.close()
        js = JSONLines(fn)
        assert convert(list, js) == dat


def test_read_gzip():
    with tmpfile("json.gz") as fn:
        f = gzip.open(fn, "wb")
        s = json.dumps(dat).encode("utf-8")
        f.write(s)
        f.close()
        js = JSON(fn)
        assert convert(list, js) == dat


def test_write_gzip_lines():
    with tmpfile("json.gz") as fn:
        j = JSONLines(fn)
        append(j, dat)

        f = gzip.open(fn)
        line = next(f)
        f.close()
        assert line.decode("utf-8").strip() == str(json.dumps(dat[0]))


def test_write_gzip():
    with tmpfile("json.gz") as fn:
        j = JSON(fn)
        append(j, dat)

        f = gzip.open(fn)
        text = f.read()
        f.close()
        assert text.decode("utf-8").strip() == str(json.dumps(dat))
        assert isinstance(resource(fn), (JSON, JSONLines))


def test_resource_gzip():
    with tmpfile("json.gz") as fn:
        assert isinstance(resource(fn), (JSON, JSONLines))
        assert isinstance(resource("json://" + fn), (JSON, JSONLines))
        assert isinstance(resource("jsonlines://" + fn), (JSON, JSONLines))

    with tmpfile("jsonlines.gz"):
        assert isinstance(resource("jsonlines://" + fn), (JSON, JSONLines))


def test_convert_to_temp_json():
    js = convert(Temp(JSON), [1, 2, 3])
    assert isinstance(js, JSON)
    assert isinstance(js, _Temp)

    assert convert(list, js) == [1, 2, 3]


def test_drop():
    with tmpfile("json") as fn:
        js = JSON(fn)
        append(js, [1, 2, 3])

        assert os.path.exists(fn)
        drop(js)
        assert not os.path.exists(fn)


def test_missing_to_csv():
    data = [dict(a=1, b=2), dict(a=2, c=4)]
    with tmpfile(".json") as fn:
        js = JSON(fn)
        js = odo(data, js)

        with tmpfile(".csv") as csvf:
            csv = odo(js, csvf)
            with open(csv.path, "rt") as f:
                result = f.read()

    expected = "a,b,c\n1,2.0,\n2,,4.0\n"
    assert result == expected



"""
    ulmo.ncdc.ghcn_daily.core
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `National Climatic Data Center`_
    `Global Historical Climate Network - Daily`_ dataset.


    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Global Historical Climate Network - Daily: http://www.ncdc.noaa.gov/oa/climate/ghcn-daily/

"""
from builtins import str
from builtins import range
from past.builtins import basestring
import itertools
import os

import numpy as np
import pandas

from tsgettoolbox.ulmo import util


GHCN_DAILY_DIR = os.path.join(util.get_ulmo_dir(), "ncdc/ghcn_daily")


def get_data(station_id, elements=None, update=True, as_dataframe=False):
    """Retrieves data for a given station.


    Parameters
    ----------
    station_id : str
        Station ID to retrieve data for.
    elements : ``None``, str, or list of str
        If specified, limits the query to given element code(s).
    update : bool
        If ``True`` (default),  new data files will be downloaded if they are
        newer than any previously cached files. If ``False``, then previously
        downloaded files will be used and new files will only be downloaded if
        there is not a previously downloaded file for a given station.
    as_dataframe : bool
        If ``False`` (default), a dict with element codes mapped to value dicts
        is returned. If ``True``, a dict with element codes mapped to equivalent
        pandas.DataFrame objects will be returned. The pandas dataframe is used
        internally, so setting this to ``True`` is a little bit faster as it
        skips a serialization step.


    Returns
    -------
    site_dict : dict
        A dict with element codes as keys, mapped to collections of values. See
        the ``as_dataframe`` parameter for more.
    """
    if isinstance(elements, basestring):
        elements = [elements]

    start_columns = [
        ("year", 11, 15, int),
        ("month", 15, 17, int),
        ("element", 17, 21, str),
    ]
    value_columns = [
        ("value", 0, 5, float),
        ("mflag", 5, 6, str),
        ("qflag", 6, 7, str),
        ("sflag", 7, 8, str),
    ]
    columns = list(
        itertools.chain(
            start_columns,
            *[
                [
                    (name + str(n), start + 13 + (8 * n), end + 13 + (8 * n), converter)
                    for name, start, end, converter in value_columns
                ]
                for n in range(1, 32)
            ]
        )
    )

    station_file_path = _get_ghcn_file(station_id + ".dly", check_modified=update)
    station_data = util.parse_fwf(station_file_path, columns, na_values=[-9999])

    dataframes = {}

    for element_name, element_df in station_data.groupby("element"):
        if not elements is None and element_name not in elements:
            continue

        element_df["month_period"] = element_df.apply(
            lambda x: pandas.Period("%s-%s" % (x["year"], x["month"])), axis=1
        )
        element_df = element_df.set_index("month_period")
        monthly_index = element_df.index

        # here we're just using pandas' builtin resample logic to construct a daily
        # index for the timespan
        # 2018/11/27 johanneshorak: hotfix to get ncdc ghcn_daily working again
        # new resample syntax requires resample method to generate resampled index.
        daily_index = element_df.resample("D").sum().index.copy()

        # XXX: hackish; pandas support for this sort of thing will probably be
        # added soon
        month_starts = (monthly_index - 1).asfreq("D") + 1
        dataframe = pandas.DataFrame(
            columns=["value", "mflag", "qflag", "sflag"], index=daily_index
        )

        for day_of_month in range(1, 32):
            dates = [
                date
                for date in (month_starts + day_of_month - 1)
                if date.day == day_of_month
            ]
            if not len(dates):
                continue
            months = pandas.PeriodIndex([pandas.Period(date, "M") for date in dates])
            for column_name in dataframe.columns:
                col = column_name + str(day_of_month)
                dataframe[column_name][dates] = element_df[col][months]

        dataframes[element_name] = dataframe

    if as_dataframe:
        return dataframes
    else:
        return dict(
            [
                (key, util.dict_from_dataframe(dataframe))
                for key, dataframe in dataframes.items()
            ]
        )


def get_stations(
    country=None,
    state=None,
    elements=None,
    start_year=None,
    end_year=None,
    update=True,
    as_dataframe=False,
):
    """Retrieves station information, optionally limited to specific parameters.


    Parameters
    ----------
    country : str
        The country code to use to limit station results. If set to ``None``
        (default), then stations from all countries are returned.
    state : str
        The state code to use to limit station results. If set to ``None``
        (default), then stations from all states are returned.
    elements : ``None``, str, or list of str
        If specified, station results will be limited to the given element codes
        and only stations that have data for any these elements will be
        returned.
    start_year : int
        If specified, station results will be limited to contain only stations
        that have data after this year. Can be combined with the ``end_year``
        argument to get stations with data within a range of years.
    end_year : int
        If specified, station results will be limited to contain only stations
        that have data before this year. Can be combined with the ``start_year``
        argument to get stations with data within a range of years.
    update : bool
        If ``True`` (default),  new data files will be downloaded if they are
        newer than any previously cached files. If ``False``, then previously
        downloaded files will be used and new files will only be downloaded if
        there is not a previously downloaded file for a given station.
    as_dataframe : bool
        If ``False`` (default), a dict with station IDs keyed to station dicts
        is returned. If ``True``, a single pandas.DataFrame object will be
        returned. The pandas dataframe is used internally, so setting this to
        ``True`` is a little bit faster as it skips a serialization step.


    Returns
    -------
    stations_dict : dict or pandas.DataFrame
        A dict or pandas.DataFrame representing station information for stations
        matching the arguments. See the ``as_dataframe`` parameter for more.
    """

    columns = [
        ("country", 0, 2, None),
        ("network", 2, 3, None),
        ("network_id", 3, 11, None),
        ("latitude", 12, 20, None),
        ("longitude", 21, 30, None),
        ("elevation", 31, 37, None),
        ("state", 38, 40, None),
        ("name", 41, 71, None),
        ("gsn_flag", 72, 75, None),
        ("hcn_flag", 76, 79, None),
        ("wm_oid", 80, 85, None),
    ]

    stations_file = _get_ghcn_file("ghcnd-stations.txt", check_modified=update)
    stations = util.parse_fwf(stations_file, columns)

    if not country is None:
        stations = stations[stations["country"] == country]
    if not state is None:
        stations = stations[stations["state"] == state]

    # set station id and index by it
    stations["id"] = stations[["country", "network", "network_id"]].T.apply("".join)

    if not elements is None or not start_year is None or not end_year is None:
        inventory = _get_inventory(update=update)
        if not elements is None:
            if isinstance(elements, basestring):
                elements = [elements]

            mask = np.zeros(len(inventory), dtype=bool)
            for element in elements:
                mask += inventory["element"] == element
            inventory = inventory[mask]
        if not start_year is None:
            inventory = inventory[inventory["last_year"] >= start_year]
        if not end_year is None:
            inventory = inventory[inventory["first_year"] <= end_year]

        uniques = inventory["id"].unique()
        ids = pandas.DataFrame(uniques, index=uniques, columns=["id"])
        stations = pandas.merge(stations, ids).set_index("id", drop=False)

    stations = stations.set_index("id", drop=False)
    # wm_oid gets convertidsed as a float, so cast it to str manually
    # pandas versions prior to 0.13.0 could use numpy's fix-width string type
    # to do this but that stopped working in pandas 0.13.0 - fortunately a
    # regex-based helper method was added then, too
    if pandas.__version__ < "0.13.0":
        stations["wm_oid"] = stations["wm_oid"].astype("|U5")
        stations["wm_oid"][stations["wm_oid"] == "nan"] = np.nan
    else:
        stations["wm_oid"] = stations["wm_oid"].astype("|U5").map(lambda x: x[:-2])
        is_nan = stations["wm_oid"] == "n"
        is_empty = stations["wm_oid"] == ""
        is_invalid = is_nan | is_empty
        stations.loc[is_invalid, "wm_oid"] = np.nan

    if as_dataframe:
        return stations
    else:
        return util.dict_from_dataframe(stations)


def _get_ghcn_file(filename, check_modified=True):
    base_url = "http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/"
    if "ghcnd-" in filename:
        url = base_url + filename
    else:
        url = base_url + "all/" + filename

    path = os.path.join(GHCN_DAILY_DIR, url.split("/")[-1])
    util.download_if_new(url, path, check_modified=check_modified)
    return path


def _get_inventory(update=True):
    columns = [
        ("id", 0, 11, None),
        ("latitude", 12, 20, None),
        ("longitude", 21, 30, None),
        ("element", 31, 35, None),
        ("first_year", 36, 40, None),
        ("last_year", 41, 45, None),
    ]

    inventory_file = _get_ghcn_file("ghcnd-inventory.txt", check_modified=update)
    return util.parse_fwf(inventory_file, columns)



#!/usr/bin/env python
r"""
tsgettoolbox command line/library tools to retrieve time series.

This program is a collection of utilities to download data from various
web services.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import warnings

from tsgettoolbox.odo import odo, resource
import pandas as pd
import mando

try:
    from mando.rst_text_formatter import RSTHelpFormatter as HelpFormatter
except ImportError:
    from argparse import RawTextHelpFormatter as HelpFormatter

from tstoolbox import tsutils

warnings.filterwarnings("ignore")


nwis_docstrings = {
    "filter_descriptions": r"""
    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and
    sites are in different time zones.  Note that the measurement time
    zone at a site may not be the same as the time zone actually in
    effect at the site.

    Every query requires a major filter. Pick the major filter
    ('--sites', '--stateCd', '--huc', '--bBox', '--countyCd') that best
    retrieves the data for the sites that you are interested in.  You
    can have only one major filter per query. If you specify more than
    one major filter, you will get an error.

    **Major Filter**

    Select ONE of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter.
    This further reduces the set of expected results. Users are
    encouraged to use minor filters because it allows more efficient use
    of this service.

    Use as many as desired to limit number of retrieved time series::

        '--agencyCd',
        '--altMax',
        '--altMin',
        '--aquiferCd',
        '--drainAreaMax',
        '--drainAreaMin',
        '--holeDepthMax'
        '--holeDepthMin',
        '--localAquiferCd',
        '--modifiedSince',
        '--parameterCd',
        '--siteStatus',
        '--siteType',
        '--wellDepthMax',
        '--wellDepthMin',
    """,
    "results_ts": r"""
    **Results**

    The column name in the resulting table is made up of
    "agencyCd_siteno_parameterCd", for example "USGS_02248380_00010".
    The agency and parameter codes are described in the `agencyCd` and
    `parameterCd` options below.

    If `includeCodes` option is used, there will also be columns representing
    the data quality codes named "agencyCd_siteno_parameterCd_cd".

    +---------+--------------------------------------------------------+
    | Code    |  Description                                           |
    +=========+========================================================+
    | e       | Value has been edited or estimated by USGS personnel   |
    |         | and is write protected                                 |
    +---------+--------------------------------------------------------+
    | &       | Value was computed from affected unit values           |
    +---------+--------------------------------------------------------+
    | E       | Value was computed from estimated unit values.         |
    +---------+--------------------------------------------------------+
    | A       | Approved for publication -- Processing and review      |
    |         | completed.                                             |
    +---------+--------------------------------------------------------+
    | P       | Provisional data subject to revision.                  |
    +---------+--------------------------------------------------------+
    | <       | The value is known to be less than reported value and  |
    |         | is write protected.                                    |
    +---------+--------------------------------------------------------+
    | >       | The value is known to be greater than reported value   |
    |         | and is write protected.                                |
    +---------+--------------------------------------------------------+
    | 1       | Value is write protected without any remark code to be |
    |         | printed                                                |
    +---------+--------------------------------------------------------+
    | 2       | Remark is write protected without any remark code to   |
    |         | be printed                                             |
    +---------+--------------------------------------------------------+
    |         | No remark (blank)                                      |
    +---------+--------------------------------------------------------+
    | Ssn     | Parameter monitored seasonally                         |
    +---------+--------------------------------------------------------+
    | Ice     | Ice affected                                           |
    +---------+--------------------------------------------------------+
    | Pr      | Partial-record site                                    |
    +---------+--------------------------------------------------------+
    | Rat     | Rating being developed or revised                      |
    +---------+--------------------------------------------------------+
    | Eqp     | Equipment malfunction                                  |
    +---------+--------------------------------------------------------+
    | Fld     | Flood damage                                           |
    +---------+--------------------------------------------------------+
    | Dis     | Data-collection discontinued                           |
    +---------+--------------------------------------------------------+
    | Dry     | Dry                                                    |
    +---------+--------------------------------------------------------+
    | --      | Parameter not determined                               |
    +---------+--------------------------------------------------------+
    | Mnt     | Maintenance in progress                                |
    +---------+--------------------------------------------------------+
    | ZFl     | Zero flow                                              |
    +---------+--------------------------------------------------------+
    | ``***`` | Temporarily unavailable                                |
    +---------+--------------------------------------------------------+
    """,
    "includeCodes": r"""includeCodes
        [optional, default is False]

        Whether or not to include the metadata/quality code column.
        Useful to almost halve the size of the pandas DataFrame.""",
    "sites": r"""sites : str
        [optional, default is None, major site filter]

        Want to only query one site? Use sites as your major filter, and
        put only one site number in the list.  Sites are comma
        separated.  Sites may be prefixed with an optional agency code
        followed by a colon. If you do not know the site numbers you
        need, you can find relevant sites with the NWIS Mapper
        (http://wdr.water.usgs.gov/nwisgmap/index.html) or on the USGS
        Water Data for the Nation site.
        (http://waterdata.usgs.gov/nwis/)

        Can have from 1 to 100 comma separated site numbers::

            --sites=USGS:01646500
            --sites=01646500,06306300""",
    "stateCd": r"""stateCd : str
        [optional, default is None, major site filter]

        U.S. postal service (2-digit) state code.  Can have only 1 state
        code.  List is available at
        http://www.usps.com/ncsc/lookups/usps_abbreviations.html::

            --stateCd=NY""",
    "huc": r"""huc : str
        [optional, default is None, major site filter]

        A list of hydrologic unit codes (HUC) or watersheds.  Only
        1 major HUC can be specified per request.  A major HUC has two
        digits. Minor HUCs must be eight digits in length.  Can have
        1 to 10 HUC codes.  List of HUCs is available at
        http://water.usgs.gov/GIS/huc_name.html::

            --huc=01,02070010""",
    "bBox": r"""bBox :
        [optional, default is None, major site filter]

        A contiguous range of decimal latitude and longitude, starting
        with the west longitude, then the south latitude, then the east
        longitude, and then the north latitude with each value separated
        by a comma. The product of the range of latitude and longitude
        cannot exceed 25 degrees. Whole or decimal degrees must be
        specified, up to six digits of precision. Minutes and seconds
        are not allowed. Remember: western longitude (which includes
        almost all of the United States) is specified in negative
        degrees. Caution: many sites outside the continental US do not
        have latitude and longitude referenced to NAD83 and therefore
        can not be found using these arguments. Certain sites are not
        associated with latitude and longitude due to homeland security
        concerns and cannot be found using this filter.::

            --bBox=-83,36.5,-81,38.5""",
    "countyCd": r"""countyCd :
        [optional, default is None, major site filter]

        A list of county numbers, in a 5 digit numeric format. The first
        two digits of a county's code are the FIPS State Code.  Can have
        from 1 to 20 county codes.  The first 2 digits are the FIPS
        State Code (http://www.itl.nist.gov/fipspubs/fip5-2.htm) and the
        list of county codes are at
        http://help.waterdata.usgs.gov/code/county_query?fmt=html::

            --countyCd=51059,51061""",
    "parameterCd": r"""parameterCd :
        [optional, default is None, minor site filter]

        USGS time-series parameter code.  All parameter codes are
        numeric and 5 characters in length.  Parameter codes are used to
        identify the constituent measured and the units of measure.
        Popular codes include stage (00065), discharge in cubic feet per
        second (00060) and water temperature in degrees Celsius (00010).
        Can request from 1 to 100 "parameterCD"s.  Default: returns all
        regular time-series for the requested sites.

        Complete list::

            http://help.waterdata.usgs.gov/codes-and-parameters/parameters::

            --parameterCd=00060       # discharge, cubic feet
                                      # per second
            --parameterCd=00060,00065 # discharge,
                                      # cubic feet per second
                                      # and gage height in
                                      # feet""",
    "siteType": r"""siteType :
        [optional, default is None, minor site filter]

        Restricts sites to those having one or more major and/or minor
        site types.  If you request a major site type (ex: &siteType=ST)
        you will get all sub-site types of the same major type as well
        (in this case, ST-CA, ST-DCH and ST-TS).  Can have from 1 to an
        unlimited number of siteType codes.  Default is to return all
        types.  List of valid site types:
        http://help.waterdata.usgs.gov/site_tp_cd::

            --siteType=ST       # Streams only
            --siteType=ST,LA-OU # Streams and Land Outcrops only""",
    "modifiedSince": r"""modifiedSince :
        [optional, default is None, minor site filter]

        Returns all values for sites and period of record requested only
        if any values have changed over the last modifiedSince period.
        modifiedSince is useful if you periodically need to poll a site
        but are only interested in getting data if some of it has
        changed.  It is typically be used with period, or startDT/endDT
        but does not have to be. In the latter case, if any values were
        changed during the specified modifiedSince period, only the most
        recent values would be retrieved for those sites. This is
        a typical usage, since users typically are polling a site and
        only want data if there are new or changed measurements.
        ISO-8601 duration format is always used.  There is no default.
        (http://en.wikipedia.org/wiki/ISO_8601#Durations)::

            --modifiedSince=PT2H
                   # Retrieves all values for sites and period of record
                   # requested for any of the requested sites and
                   # parameters, but only for sites where any of the
                   # values changed during the last two hours.
            --modifiedSince=PT2H --period=P1D
                   # Retrieve all values for sites and period of record
                   # requested for the last 24 hours from now only for
                   # sites and parameters that had any values that
                   # changed or were added during the last two hours.
            --modifiedSince=PT2H --startDt=2010-11-01 --endDt=2010-11-02
                   # Retrieve all values for sites and period of record
                   # requested for sites and parameters that had values
                   # change between midnight site local time on Nov 1st,
                   # 2010 and 23:59 on Nov 2nd, 2010 site local time,
                   # only if values were changed or added within the
                   # last two hours.""",
    "agencyCd": r"""agencyCd :
        [optional, default is None, minor site filter]

        The list of sites returned are filtered to return only those
        with the provided agency code. The agency code describes the
        organization that maintains the site. Only one agency code is
        allowed and is optional.  An authoritative list of agency codes
        can be found here.  Default is to return all sites regardless of
        agency code.  List:
        http://help.waterdata.usgs.gov/code/agency_cd_query?fmt=html::

            --stateCd=il --agencyCd=USCE # Only US Army Corps
                                         # of Engineers sites
                                         # in Illinois""",
    "siteStatus": r"""siteStatus :
        [optional, default is None, minor site filter]

        Selects sites based on whether or not they are active. If a site
        is active, it implies that it is being actively maintained.
        A site is considered active if: it has collected time-series
        (automated) data within the last 183 days (6 months), or it has
        collected discrete (manually collected) data within 397 days (13
        months) If it does not meet these criteria, it is considered
        inactive. Some exceptions apply.  If a site is flagged by a USGS
        water science center as discontinued, it will show as inactive.
        A USGS science center can also flag a new site as active even if
        it has not collected any data.  The default is all (show both
        active and inactive sites).  Chose between, 'all', 'active', or
        'inactive'.  Default all - sites of any activity status are
        returned.::

            --siteStatus='active'""",
    "altMin": r"""altMin : float
        [optional, default is None, minor site filter]

        These arguments allows you to select instantaneous values sites
        where the associated sites' altitude are within a desired
        altitude, expressed in feet.  Altitude is based on the datum
        used at the site.  Providing a value to altMin (minimum
        altitude) means you want sites that have or exceed the altMin
        value.  You may specify decimal feet if precision is critical If
        both the altMin and altMax are specified, sites at or between
        the minimum and maximum altitude are returned.""",
    "altMax": r"""altMax : float
        [optional, default is None, minor site filter]

        Providing a value to altMax (maximum altitude) means you want
        sites that have or are less than the altMax value.::

            --altMin=1000 --altMax=5000
                  # Return sites where the altitude is 1000 feet or
                  # greater and 5000 feet or less.
            --altMin=12.5 --altMax=13
                  # Return sites where the altitude is 12.5 feet or
                  # greater and 13 feet or less.""",
    "drainAreaMin": r"""drainAreaMin : float
        [optional, default is None, minor site filter]

        SURFACE WATER SITE ATTRIBUTE

        These arguments allows you to select principally surface water
        sites where the associated sites' drainage areas (watersheds)
        are within a desired size, expressed in square miles or decimal
        fractions thereof.  Providing a value to drainAreaMin (minimum
        drainage area) means you want sites that have or exceed the
        drainAreaMin value.  The values may be expressed in decimals. If
        both the drainAreaMin and drainAreaMax are specified, sites at
        or between the minimum and maximum drainage areas values
        specified are returned Caution: not all sites are associated
        with a drainage area.  Caution: drainage area generally only
        applies to surface water sites.  Use with other site types, such
        as groundwater sites, will likely retrieve no results.""",
    "drainAreaMax": r"""drainAreaMax:  float
        [optional, default is None, minor site filter]

        SURFACE WATER SITE ATTRIBUTE

        Providing a value to drainAreaMax (maximum drainage area) means
        you want sites that have or are less than the drainAreaMax
        value.::

            --drainAreaMin=1000 --drainAreaMax=5000
                                 # Return sites where the drainage area
                                 # is 1000 square miles or greater and
                                 # is 5000 square miles or less.
            --drainAreaMin=10.5 --drainAreaMax=10.7
                                 # Return sites where the drainage area
                                 # is 10.5 square miles or greater and
                                 # is 10.7 square miles or less.""",
    "aquiferCd": r"""aquiferCd
        [optional, default is None, minor site filter]

        Used to filter sites to those that exist in specified national
        aquifers. Note: not all sites have been associated with national
        aquifers.  Enter one or more national aquifer codes, separated
        by commas.  A national aquifer code is exactly 10 characters.
        You can have up to 1000 aquiferCd codes.  Complete list:
        http://water.usgs.gov/ogw/NatlAqCode-reflist.html::

            --aquiferCd=S500EDRTRN,N100HGHPLN
                                  # returns groundwater sites for the
                                  # Edwards-Trinity aquifer system and
                                  # the High Plains national
                                  # aquifers.""",
    "localAquiferCd": r"""localAquiferCd
        [optional, default is None, minor site filter]

        Used to filter sites to those that exist in specified local
        aquifers.  Note: not all sites have been associated with local
        aquifers.  Enter one or more local aquifer codes, separated by
        commas.  A local aquifer code begins with a 2 character state
        abbreviation (such as TX for Texas) followed by a colon followed
        by the 7 character aquifer code.  Can have 0 to 1000 comma
        delimited codes.  Complete list:
        http://help.waterdata.usgs.gov/code/aqfr_cd_query?fmt=html
        To translate state codes associated with the local aquifer you
        may need this reference:
        http://www.itl.nist.gov/fipspubs/fip5-2.htm ::

            --localAquiferCd=AL:111RGLT,AL:111RSDM
                    # returns sites for the Regolith and Saprolite local
                    # aquifers in Alabama""",
    "wellDepthMin": r"""wellDepthMin : float
        [optional, default is None, minor site filter]

        GROUNDWATER SITE ATTRIBUTE

        These arguments allows you to select groundwater sites serving
        data recorded automatically where the associated sites' well
        depth are within a desired depth, expressed in feet from the
        land surface datum.  Express well depth as a positive number.
        Providing a value to wellDepthMin (minimum well depth) means you
        want sites that have or exceed the wellDepthMin value.  The
        values may be expressed in decimals Caution: well depth applies
        to groundwater sites only.::

             --wellDepthMin=100 --wellDepthMax=500
                     # Return daily value sites where the well depth is
                     # 100 feet or greater and 500 feet or less.""",
    "wellDepthMax": r"""wellDepthMax : float
        [optional, default is None, minor site filter]

        GROUNDWATER SITE ATTRIBUTE

        Providing a value to wellDepthMax (maximum well depth) means you
        want sites that have or are less than the wellDepthMax value.::

             --wellDepthMin=10.5 --wellDepthMax=10.7
                     # Return daily value sites where the well depth is
                     # 10.5 feet or greater and 10.7 feet or less.

        If both the wellDepthMin and wellDepthMax are specified, sites
        at or between the minimum and maximum well depth values
        specified are returned wellDepthMax should be greater than or
        equal to wellDepthMin.""",
    "holeDepthMin": r"""holeDepthMin : float
        [optional, default is None, minor site filter]

        GROUNDWATER SITE ATTRIBUTE

        These arguments allows you to select groundwater sites serving
        data recorded automatically where the associated sites' hole
        depth are within a desired depth, expressed in feet from the
        land surface datum.  Express hole depth as a positive number.
        Providing a value to holeDepthMin (minimum hole depth) means you
        want sites that have or exceed the holeDepthMin value.  The
        values may be expressed in decimals Caution: hole depth applies
        to groundwater sites only.""",
    "holeDepthMax": r"""holeDepthMax : float
        [optional, default is None, minor site filter]

        GROUNDWATER SITE ATTRIBUTE

        Providing a value to holeDepthMax (maximum hole depth) means you
        want sites that have or are less than the holeDepthMax value.::

            --holeDepthMin=100 --holeDepthMax=500
                    # Return daily values sites where the hole depth is
                    # 100 feet or greater and 500 feet or less.

            --holeDepthMin=10.5 --holeDepthMax=10.7
                    # Return daily value sites where the hole depth is
                    # 10.5 feet or greater and 10.7 feet or less.

        If both the holeDepthMin and holeDepthMax are specified, sites
        at or between the minimum and maximum hole depth values
        specified are returned holeDepthMax should be greater than or
        equal to holeDepthMin.""",
    "period": r"""period
        [optional, default is None]

        Get a range of values from now by specifying the period argument
        period must be in ISO-8601 Duration format.
        (http://en.wikipedia.org/wiki/ISO_8601#Durations) Negative
        periods (ex: P-T2H) are not allowed.  Data are always returned
        up to the most recent value, which in the case of a predictive
        gage might be in the future.  When specifying days from now, the
        first value will probably not be at midnight of the first day,
        but somewhat before exactly 24 hours from now.::

            --period=PT2H
                  # Retrieve last two hours from now up to most recent
                  # instantaneous value)
            --period=P7D
                  # Retrieve last seven days up from now to most recent
                  # instantaneous value)""",
    "startDT": r"""startDT
        [optional, default is None]

        Get a range of values from an explicit begin or end date/time.
        Use the startDT and endDT arguments.  Site local time is output,
        even if multiple sites are requested and sites are in different
        time zones.  Note that the measurement time zone at a site may
        not be the same as the time zone actually in effect at the site.

        Both startDt and endDt must be in ISO-8601 Date/Time format.
        (http://en.wikipedia.org/wiki/ISO_8601#Dates) You can express
        the date and time in a timezone other than site local time if
        you want as long as it follows the ISO standard. For example,
        you can express the time in Universal time: 2014-03-20T00:00Z.
        If startDT is supplied and endDT is not, endDT ends with the
        most recent instantaneous value. startDT must be chronologically
        before endDT.

        If startDt shows the date and not the time of day (ex:
        2010-09-01) the time of midnight site time is assumed
        (2010-09-01T00:00) If endDt shows the date and not the time of
        day (ex: 2010-09-02) the last minute before midnight site time
        is assumed (2010-09-02T23:59).  Remember, only data from October
        1, 2007 are currently available in the 'iv' database.""",
    "endDT": r"""endDT
        [optional, default is None]

        If endDT is present, startDt must also be
        present.::

            --startDT=2010-11-22 --endDT=2010-11-22  # Full day, 00:00 to 23:59
            --startDT=2010-11-22T12:00 --endDT=2010-11-22T18:00
            --startDT=2010-11-22 --endDT=2010-11-22
            --startDT=2010-11-22T12:00  # From "startDT" to most recent
                                        # instantaneous value""",
    "statReportType": r"""statReportType : str
        [optional, default is 'daily']

        The type of statistics desired. Valid statistic report types
        include:

        +----------------+------------------------------------------+
        | statReportType | Description                              |
        +----------------+------------------------------------------+
        | daily          | daily statistics (default)               |
        |                | statistic across years                   |
        +----------------+------------------------------------------+
        | monthly        | monthly statistics (monthly time-series) |
        +----------------+------------------------------------------+
        | annual         | annual statistics, based on either       |
        |                | calendar year or water year, as defined  |
        |                | by statYearType. If statYearType is not  |
        |                | provided, calendar year statistics are   |
        |                | assumed. (annual time-series)            |
        +----------------+------------------------------------------+""",
    "statType": r"""statType : str
        [optional, default is None, minor site filter]

        Selects sites based on the statistics type(s) desired, such as
        minimum, maximum or mean

        For all statReportType types include::

            mean - arithmetic mean or average
            all - selects all available statistics

        For daily statistics you can also specify::

            min - minimum, or smallest value found for the
                  daily statistics
            max - maximum, or largest value found for the
                  daily statistics
            median - the numerical value separating the higher
                     half of a the data from the lower half,
                     same as specifying P50. If used median
                     will be represented by the column name
                     p50_va.
            P05, P10, P20, P25, P50, P75, P80, P90, P95
                with the number indicating percentile. Note:
                the service can calculate only these
                percentiles.""",
    "missingData": r"""missingData
        [optional, default is None]

        Used to indicate the rules to follow to generate statistics if
        there are gaps in the period of record during the requested
        statistics period. By default if there are any missing data for
        the report type, the statistic is left blank or null.

        This option does not apply to daily statistics, but optionally
        can be used with monthly and yearly statistics. If used with
        daily statistics, an error will occur.

        Missing data can happen for various reasons including there was
        a technical problem with the gage for part of the time period.

        Enabling this switch will attempt to provide a statistic if
        there is enough data to create one.

        Choice is 'off' or 'on'.""",
    "statisticsCd": r"""statisticsCd
        [optional, default is None]

        The statisticsCd represents how the instantaneous values are
        aggregated.  The statisticsCd is from the following table:

        +-------+------------------------------------+
        | Code  | Description                        |
        +=======+====================================+
        | 00001 | MAXIMUM VALUES                     |
        +-------+------------------------------------+
        | 00002 | MINIMUM VALUES                     |
        +-------+------------------------------------+
        | 00003 | MEAN VALUES                        |
        +-------+------------------------------------+
        | 00004 | VALUES TAKEN BETWEEN 0001 AND 1200 |
        +-------+------------------------------------+
        | 00005 | VALUES TAKEN BETWEEN 1201 AND 2400 |
        +-------+------------------------------------+
        | 00006 | SUMMATION VALUES                   |
        +-------+------------------------------------+
        | 00007 | MODAL VALUES                       |
        +-------+------------------------------------+
        | 00008 | MEDIAN VALUES                      |
        +-------+------------------------------------+
        | 00009 | STANDARD DEVIATION VALUES          |
        +-------+------------------------------------+
        | 00010 | VARIANCE VALUES                    |
        +-------+------------------------------------+
        | 00011 | RANDOM INSTANTANEOUS VALUES        |
        +-------+------------------------------------+
        | 00012 | EQUIVALENT MEAN VALUES             |
        +-------+------------------------------------+
        | 00013 | SKEWNESS VALUES                    |
        +-------+------------------------------------+
        | 00021 | TIDAL HIGH-HIGH VALUES             |
        +-------+------------------------------------+
        | 00022 | TIDAL LOW-HIGH VALUES              |
        +-------+------------------------------------+
        | 00023 | TIDAL HIGH-LOW VALUES              |
        +-------+------------------------------------+
        | 00024 | TIDAL LOW-LOW VALUES               |
        +-------+------------------------------------+
        | 01XXY | XX.Y PERCENTILE                    |
        +-------+------------------------------------+
        | 02LLL | LLL DAY LOW MEAN                   |
        +-------+------------------------------------+
        | 03HHH | HHH DAY HIGH MEAN                  |
        +-------+------------------------------------+
        | 3TTTT | INSTANTANEOUS OBSERVATION AT TTTT  |
        +-------+------------------------------------+""",
    "siteOutput": r"""siteOutput
        [optional, default is None]

        If you would like to see expanded site information, check this
        box.  This argument is ignored for visually oriented output
        formats like Mapper, Google Earth and Google Maps. The default
        is basic. Use expanded to get expanded site information.
        Example: &siteOutput=expanded. Note: for performance reasons,
        &siteOutput=expanded cannot be used if seriesCatalogOutput=true
        or with any values for outputDataTypeCd.""",
    "seriesCatalogOutput": r"""seriesCatalogOutput
        [optional, default is None]

        This argument is ignored for visually oriented output formats
        like Mapper, Google Earth and Google Maps. If you would like to
        see all the period of record information for the sites selected,
        check this box.  You will see detailed information, such as
        a continuous range of dates served by a site for one or more
        data types, for example, the begin and end dates that streamflow
        (parameter 00060) was recorded at a site.  Note: if you select
        any data types for output (see below) the period of record data
        will also appear. In that case specifying this argument is
        unnecessary. The default is false. The only legal values for
        this argument are true and false. Example:
        &seriesCatalogOutput=true.
        &seriesCatalogOutput=true is equivalent to
        &outputDataTypeCd=all. Note: for performance reasons,
        &siteOutput=expanded cannot be used if
        seriesCatalogOutput=true.""",
    "outputDataTypeCd": r"""outputDataTypeCd
        [optional, default is None]

        This will add period of record information to certain output
        formats (GML, RDB and JSON) that summarize information about the
        data types requested.  The default is all data types. Some
        output formats are designed for visual use (Google Earth, Google
        Maps and Mapper).  Consequently with these formats you will not
        see data type code information.

        Default information: If seriesCatalogOutput is true, all period
        of record information is shown by default. If
        seriesCatalogOutput is false, unless you override it using one
        of the values below, no period of record information is shown.

        Note: for performance reasons, &siteOutput=expanded cannot be
        used if with any values for outputDataTypeCd.

        Here are the various output data type codes available. These can
        be selected individually or can be added as comma separated
        values if desired.  Example: &outputDataTypeCd=iv,dv

        +-----+---------------------------------------------------------------+
        | all | default (see above for qualifications). This is equivalent to |
        |     | &seriesCatalogOutput=true.                                    |
        +-----+---------------------------------------------------------------+
        | iv  | Instantaneous values (time-series measurements typically      |
        |     | recorded by automated equipment at frequent intervals (e.g.,  |
        |     | hourly)                                                       |
        +-----+---------------------------------------------------------------+
        | uv  | Unit values (alias for iv)                                    |
        +-----+---------------------------------------------------------------+
        | rt  | Real-time data (alias for iv)                                 |
        +-----+---------------------------------------------------------------+
        | dv  | Daily values (once daily measurements or summarized           |
        |     | information for a particular day, such as daily maximum,      |
        |     | minimum and mean)                                             |
        +-----+---------------------------------------------------------------+
        | pk  | Peaks measurements of water levels and streamflow for surface |
        |     | water sites (such as during floods, may be either an          |
        |     | automated or a manual measurement)                            |
        +-----+---------------------------------------------------------------+
        | sv  | Site visits (irregular manual surface water measurements,     |
        |     | excluding peak measurements)                                  |
        +-----+---------------------------------------------------------------+
        | gw  | Groundwater levels measured at irregular, discrete intervals. |
        |     | For recorded, time series groundwater levels, use iv or id.   |
        +-----+---------------------------------------------------------------+
        | qw  | Water-quality data from discrete sampling events and analyzed |
        |     | in the field or in a laboratory. For recorded time series     |
        |     | water-quality data, use iv or id.                             |
        +-----+---------------------------------------------------------------+
        | id  | Historical instantaneous values (sites in the USGS            |
        |     | Instantaneous Data Archive External Link)                     |
        +-----+---------------------------------------------------------------+
        | aw  | Sites monitored by the USGS Active Groundwater Level Network  |
        |     | External Link                                                 |
        +-----+---------------------------------------------------------------+
        | ad  | Sites included in USGS Annual Water Data Reports External     |
        |     | Link}                                                         |
        +-----+---------------------------------------------------------------+""",
    "siteName": r"""siteName
        [optional, default is None, minor site filter]

        This filter allows you to find a site by its name, using either
        the exact site name or a partial site name. Note that a major
        filter is still required. String matches are case insensitive,
        so if you specify "Boulder" you will retrieve site names with
        "Boulder", "boulder", "BOULDER" as well as many other variants.
        To embed a space, you can use single quotes. Examaple:
        --siteName='Boulder Creek'""",
    "siteNameMatchOperator": r"""siteNameMatchOperator
        [optional, default is None, minor site filter]

        If used, this must be used with siteName. It determines how the
        pattern matching for the site name behaves. Matches are case
        insensitive. The options are::

            start = The string must be at the start of the site name (default)
            any = The string must be contained somewhere in the site name
            exact = The site name must exactly match the string supplied, with
                    the exception that the match is not case sensitive

        Example: &siteNameMatchOperator=any""",
    "hasDataTypeCd": r"""hasDataTypeCd
        [optional, default is None, minor site filter]

        Default is all. Restricts results to those sites that collect
        certain kinds of data. Separate values with commas. Allowed
        values are:

        +-----+---------------------------------------------------------------+
        | all | default (see above for qualifications). This is equivalent to |
        |     | &seriesCatalogOutput=true.                                    |
        +-----+---------------------------------------------------------------+
        | iv  | Instantaneous values (time-series measurements typically      |
        |     | recorded by automated equipment at frequent intervals (e.g.,  |
        |     | hourly)                                                       |
        +-----+---------------------------------------------------------------+
        | uv  | Unit values (alias for iv)                                    |
        +-----+---------------------------------------------------------------+
        | rt  | Real-time data (alias for iv)                                 |
        +-----+---------------------------------------------------------------+
        | dv  | Daily values (once daily measurements or summarized           |
        |     | information for a particular day, such as daily maximum,      |
        |     | minimum and mean)                                             |
        +-----+---------------------------------------------------------------+
        | pk  | Peaks measurements of water levels and streamflow for surface |
        |     | water sites (such as during floods, may be either an          |
        |     | automated or a manual measurement)                            |
        +-----+---------------------------------------------------------------+
        | sv  | Site visits (irregular manual surface water measurements,     |
        |     | excluding peak measurements)                                  |
        +-----+---------------------------------------------------------------+
        | gw  | Groundwater levels measured at irregular, discrete intervals. |
        |     | For recorded, time series groundwater levels, use iv or id.   |
        +-----+---------------------------------------------------------------+
        | qw  | Water-quality data from discrete sampling events and analyzed |
        |     | in the field or in a laboratory. For recorded time series     |
        |     | water-quality data, use iv or id.                             |
        +-----+---------------------------------------------------------------+
        | id  | Historical instantaneous values (sites in the USGS            |
        |     | Instantaneous Data Archive External Link)                     |
        +-----+---------------------------------------------------------------+
        | aw  | Sites monitored by the USGS Active Groundwater Level Network  |
        |     | External Link                                                 |
        +-----+---------------------------------------------------------------+
        | ad  | Sites included in USGS Annual Water Data Reports External     |
        |     | Link                                                          |
        +-----+---------------------------------------------------------------+""",
    "statYearType": r"""statYearType
        [optional, default is None]

        Indicates which kind of year statistics should be created
        against. This only applies when requesting annual statistics,
        i.e.  statReportType=annual. Valid year types codes include:

        +----------+----------------------------------------------------------+
        | calendar | calendar year, i.e. January 1 through December 31        |
        +----------+----------------------------------------------------------+
        | water    | water year, i.e. a year begins October 1 of the previous |
        |          | year and ends September 30 of the current year. This is  |
        |          | the same as a federal fiscal year.                       |
        +----------+----------------------------------------------------------+""",
}


@mando.command("nwis", formatter_class=HelpFormatter, doctype="numpy")
def nwis_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    database="dv",
    statReportType=None,
    statType=None,
    missingData=None,
    statYearType=None,
):
    r"""Deprecated: Use the ``nwis_*`` functions instead.

    This "nwis" function has been split up into individual functions for each
    source database.  This allows for keywords and output to be tailored to
    each specific web service.

    +-------------------+-------------------------------+
    | New Function      | Database Name and Description |
    +===================+===============================+
    | nwis_iv           | Instantaneous Value sub-daily |
    |                   | interval data starting in     |
    |                   | 2007                          |
    +-------------------+-------------------------------+
    | nwis_dv           | Daily Values database         |
    +-------------------+-------------------------------+
    | nwis_stat         | daily/monthly/annual          |
    |                   | statistics                    |
    +-------------------+-------------------------------+
    | nwis_site         | Site metadata                 |
    +-------------------+-------------------------------+
    | nwis_measurements | Field measurements            |
    +-------------------+-------------------------------+
    | nwis_peak         | Peak flow and stage           |
    +-------------------+-------------------------------+
    | nwis_gwlevels     | Ground water levels           |
    +-------------------+-------------------------------+

    This function/sub-command will continue to work, however you should
    change all scripts to use the split out functions.
    """
    tsutils._printiso(
        nwis(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            period=period,
            startDT=startDT,
            endDT=endDT,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            drainAreaMin=drainAreaMin,
            drainAreaMax=drainAreaMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            database=database,
            statReportType=statReportType,
            statType=statType,
            missingData=missingData,
            statYearType=statYearType,
        )
    )


def nwis(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    database="dv",
    statReportType=None,
    statType=None,
    missingData=None,
    statYearType=None,
):
    warnings.warn(tsutils.error_wrapper(nwis_cli.__doc__))

    from tsgettoolbox.services.usgs import nwis as placeholder

    if database not in ["iv", "dv", "stat", "measurements", "peak", "site", "gwlevels"]:
        raise ValueError(
            tsutils.error_wrapper(
                r"""
The 'database' option must be either 'iv' for instantaneous values, or 'dv' for
daily values, or 'stat' for daily, monthly, or annual statistics, or
'measurements' for field measurements, 'peak' for peak stage and flow
estimates, 'site' for site metadata, or 'gwlevels' for ground water levels.

You gave {0}.""".format(
                    database
                )
            )
        )
    url = r"http://waterservices.usgs.gov/nwis/{0}/".format(database)
    if database in ["measurements", "peak", "gwlevels"]:
        words = sites.split(",")
        if len(words) != 1:
            raise ValueError(
                tsutils.error_wrapper(
                    r"""
For the 'measurements', 'peak', and 'gwlevels' databases you can only collect
data from one site, you listed {0}.
""".format(
                        len(words)
                    )
                )
            )
        if (
            stateCd is not None
            or huc is not None
            or bBox is not None
            or countyCd is not None
        ):
            raise ValueError(
                tsutils.error_wrapper(
                    r"""
The 'measurements', 'peak', or 'gwlevels' databases can currently only
accept one site using the 'site' keyword.
"""
                )
            )

        if database in ["measurements", "peak"]:
            url = r"http://nwis.waterdata.usgs.gov/XX/nwis/{0}".format(database)
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        drainAreaMin=drainAreaMin,
        drainAreaMax=drainAreaMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
        statReportType=statReportType,
        statType=statType,
        missingData=missingData,
        statYearType=statYearType,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_iv", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_iv_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    period=None,
    startDT=None,
    endDT=None,
    includeCodes=False,
):
    r"""Download from Instantaneous Values of the USGS NWIS.
    {filter_descriptions}
    {results_ts}
    Parameters
    ----------
    {sites}
    {stateCd}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {altMax}
    {altMin}
    {aquiferCd}
    {drainAreaMax}
    {drainAreaMin}
    {holeDepthMax}
    {holeDepthMin}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {siteStatus}
    {siteType}
    {wellDepthMax}
    {wellDepthMin}
    {period}
    {startDT}
    {endDT}
    {includeCodes}"""
    tsutils._printiso(
        nwis_iv(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            drainAreaMin=drainAreaMin,
            drainAreaMax=drainAreaMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
            includeCodes=includeCodes,
        )
    )


def nwis_iv(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    includeCodes=False,
):
    r"""Download from Instantaneous Values of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://waterservices.usgs.gov/nwis/iv/"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        drainAreaMin=drainAreaMin,
        drainAreaMax=drainAreaMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
        includeCodes=includeCodes,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_dv", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_dv_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    agencyCd=None,
    altMax=None,
    altMin=None,
    aquiferCd=None,
    drainAreaMax=None,
    drainAreaMin=None,
    holeDepthMax=None,
    holeDepthMin=None,
    localAquiferCd=None,
    modifiedSince=None,
    parameterCd=None,
    siteStatus=None,
    siteType=None,
    wellDepthMax=None,
    wellDepthMin=None,
    startDT=None,
    endDT=None,
    period=None,
    includeCodes=False,
    statisticsCd=None,
):
    r"""Download from the Daily Values database of the USGS NWIS.
    {filter_descriptions}
    {results_ts}
    Parameters
    ----------
    {sites}
    {stateCd}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {altMax}
    {altMin}
    {aquiferCd}
    {drainAreaMax}
    {drainAreaMin}
    {holeDepthMax}
    {holeDepthMin}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {siteStatus}
    {siteType}
    {wellDepthMax}
    {wellDepthMin}
    {period}
    {startDT}
    {endDT}
    {includeCodes}
    {statisticsCd}"""
    tsutils._printiso(
        nwis_dv(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            statisticsCd=statisticsCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            drainAreaMin=drainAreaMin,
            drainAreaMax=drainAreaMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
            includeCodes=includeCodes,
        )
    )


def nwis_dv(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    statisticsCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    includeCodes=False,
):
    r"""Download from the Daily Values database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://waterservices.usgs.gov/nwis/dv/"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        statisticsCd=statisticsCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        drainAreaMin=drainAreaMin,
        drainAreaMax=drainAreaMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
        includeCodes=includeCodes,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_site", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_site_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    siteOutput=None,
    seriesCatalogOutput=None,
    outputDataTypeCd=None,
    siteName=None,
    siteNameMatchOperator=None,
    hasDataTypeCd=None,
):
    r"""Download from the site database of the USGS NWIS.

    This does not return a time-series, but a table of sites.
    {filter_descriptions}
    **Results**

    +---------------------+----------------------------------+
    | Column              | Description                      |
    +=====================+==================================+
    | agency_cd           | Agency                           |
    +---------------------+----------------------------------+
    | site_no             | Site identification number       |
    +---------------------+----------------------------------+
    | station_nm          | Site name                        |
    +---------------------+----------------------------------+
    | site_tp_cd          | Site type                        |
    +---------------------+----------------------------------+
    | dec_lat_va          | Decimal latitude                 |
    +---------------------+----------------------------------+
    | dec_long_va         | Decimal longitude                |
    +---------------------+----------------------------------+
    | coord_acy_cd        | Latitude-longitude accuracy      |
    +---------------------+----------------------------------+
    | dec_coord_datum_cd  | Decimal Latitude-longitude datum |
    +---------------------+----------------------------------+
    | alt_va              | Altitude of Gage/land surface    |
    +---------------------+----------------------------------+
    | alt_acy_va          | Altitude accuracy                |
    +---------------------+----------------------------------+
    | alt_datum_cd        | Altitude datum                   |
    +---------------------+----------------------------------+
    | huc_cd              | Hydrologic unit code             |
    +---------------------+----------------------------------+

    .. _site_tp_cd: https://help.waterdata.usgs.gov/code/site_tp_query?fmt=html
    .. _coord_acy_cd: https://help.waterdata.usgs.gov/code/coord_acy_cd_query?fmt=html
    .. _dec_coord_datum_cd: https://help.waterdata.usgs.gov/code/coord_datum_cd_query?fmt=html
    .. _alt_datum_cd: https://help.waterdata.usgs.gov/code/alt_datum_cd_query?fmt=html
    .. _huc_cd: https://help.waterdata.usgs.gov/code/hucs_query?fmt=html

    Parameters
    ----------
    {sites}
    {stateCd}
    {huc}
    {bBox}
    {countyCd}

    {parameterCd}
    {siteType}
    {modifiedSince}
    {agencyCd}
    {siteStatus}
    {altMin}
    {altMax}
    {drainAreaMin}
    {drainAreaMax}
    {aquiferCd}
    {localAquiferCd}
    {wellDepthMin}
    {wellDepthMax}
    {holeDepthMin}
    {holeDepthMax}

    {period}
    {startDT}
    {endDT}
    {includeCodes}

    {siteOutput}
    {seriesCatalogOutput}
    {outputDataTypeCd}
    {siteName}
    {siteNameMatchOperator}
    {hasDataTypeCd}"""

    tsutils._printiso(
        nwis_site(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            drainAreaMin=drainAreaMin,
            drainAreaMax=drainAreaMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
            siteOutput=siteOutput,
            seriesCatalogOutput=seriesCatalogOutput,
            outputDataTypeCd=outputDataTypeCd,
            siteName=siteName,
            siteNameMatchOperator=siteNameMatchOperator,
            hasDataTypeCd=hasDataTypeCd,
        )
    )


def nwis_site(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    siteOutput=None,
    seriesCatalogOutput=None,
    outputDataTypeCd=None,
    siteName=None,
    siteNameMatchOperator=None,
    hasDataTypeCd=None,
):
    r"""Download from the site database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://waterservices.usgs.gov/nwis/site/"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        drainAreaMin=drainAreaMin,
        drainAreaMax=drainAreaMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
        siteOutput=siteOutput,
        seriesCatalogOutput=seriesCatalogOutput,
        outputDataTypeCd=outputDataTypeCd,
        siteName=siteName,
        siteNameMatchOperator=siteNameMatchOperator,
        hasDataTypeCd=hasDataTypeCd,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_gwlevels", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_gwlevels_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    r"""Download from the Groundwater Levels database of the USGS NWIS.

    {filter_descriptions}
    **Results**

    +---------------+-------------------------------+
    | Column        | Description                   |
    +===============+===============================+
    | agency_cd     | Agency code                   |
    +---------------+-------------------------------+
    | site_no       | USGS site number              |
    +---------------+-------------------------------+
    | site_tp_cd    | Site type code                |
    +---------------+-------------------------------+
    | lev_dt        | Date level measured           |
    +---------------+-------------------------------+
    | lev_tm        | Time level measured           |
    +---------------+-------------------------------+
    | lev_tz_cd     | Time datum                    |
    +---------------+-------------------------------+
    | lev_va        | Water-level value in feet     |
    |               | below land surface            |
    +---------------+-------------------------------+
    | sl_lev_va     | Water-level value in feet     |
    |               | above specific vertical datum |
    +---------------+-------------------------------+
    | sl_datum_cd   | Referenced vertical datum     |
    +---------------+-------------------------------+
    | lev_status_cd | Status                        |
    +---------------+-------------------------------+
    | lev_agency_cd | Measuring agency              |
    +---------------+-------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(
        nwis_gwlevels(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
        )
    )


def nwis_gwlevels(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://waterservices.usgs.gov/nwis/gwlevels/"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_measurements", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_measurements_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    r"""Download from the Measurements database of the USGS NWIS.

    {filter_descriptions}
    **Results**

    +---------------------------+-------------------------------------------+
    | Column                    | Description                               |
    +===========================+===========================================+
    | agency_cd                 | Agency code                               |
    +---------------------------+-------------------------------------------+
    | site_no                   | Site number                               |
    +---------------------------+-------------------------------------------+
    | measurement_nu            | Measurement number                        |
    +---------------------------+-------------------------------------------+
    | measurement_dt            | date of measurement (format = MMDDYYYY    |
    |                           | or Month/Day/Year. The user has options   |
    |                           | for the data output format)               |
    +---------------------------+-------------------------------------------+
    | tz_cd                     | Time zone offset. An ANSI SQL/92 time     |
    |                           | zone offset string. Some examples are     |
    |                           | '-07:00' (Eastern), '+02:00' (Eastern     |
    |                           | Europe), and '+03:30' (India).            |
    +---------------------------+-------------------------------------------+
    | q_meas_used_fg            | Flag indicates if the discharge           |
    |                           | measurement is marked used.               |
    +---------------------------+-------------------------------------------+
    | party_nm                  | an indication of who made the             |
    |                           | measurement and is usually populated with |
    |                           | a pair of initials separated with a slash |
    +---------------------------+-------------------------------------------+
    | site_visit_coll_agency_cd | What agency made the measurement at the   |
    |                           | site visit                                |
    +---------------------------+-------------------------------------------+
    | gage_height_va            | gage height as shown on the inside staff  |
    |                           | gage at the site or read off the recorder |
    |                           | inside the gage house in feet             |
    +---------------------------+-------------------------------------------+
    | discharge_va              | the computed discharge in cubic feet per  |
    |                           | second (cfs)                              |
    +---------------------------+-------------------------------------------+
    | measured_rating_diff      | measurement rating codes that denote the  |
    |                           | relative quality of the measurement       |
    +---------------------------+-------------------------------------------+
    | gage_va_change            | The amount the gage height changed while  |
    |                           | the measurement was being made in feet    |
    +---------------------------+-------------------------------------------+
    | gage_va_time              | The amount of time elapsed while the      |
    |                           | measurement was being made in decimal     |
    |                           | hours                                     |
    +---------------------------+-------------------------------------------+
    | control_type_cd           | condition of the rating control at the    |
    |                           | time of the measurement                   |
    +---------------------------+-------------------------------------------+
    | discharge_cd              | The adjustment code for the measured      |
    |                           | discharge                                 |
    +---------------------------+-------------------------------------------+
    | chan_nu                   | The channel number                        |
    +---------------------------+-------------------------------------------+
    | chan_name                 | The channel name                          |
    +---------------------------+-------------------------------------------+
    | meas_type                 | The channel measurement type              |
    +---------------------------+-------------------------------------------+
    | streamflow_method         | The channel discharge measurement method  |
    +---------------------------+-------------------------------------------+
    | velocity_method           | The channel velocity measurement method   |
    +---------------------------+-------------------------------------------+
    | chan_discharge            | The channel discharge in cubic feet per   |
    |                           | second                                    |
    +---------------------------+-------------------------------------------+
    | chan_width                | The channel width in feet                 |
    +---------------------------+-------------------------------------------+
    | chan_area                 | The channel area in square feet           |
    +---------------------------+-------------------------------------------+
    | chan_velocity             | The mean velocity in feet per second      |
    +---------------------------+-------------------------------------------+
    | chan_stability            | The stability of the channel material     |
    +---------------------------+-------------------------------------------+
    | chan_material             | The channel material                      |
    +---------------------------+-------------------------------------------+
    | chan_evenness             | The channel evenness from bank to bank    |
    +---------------------------+-------------------------------------------+
    | long_vel_desc             | The longitudinal velocity description     |
    +---------------------------+-------------------------------------------+
    | horz_vel_desc             | The horizontal velocity description       |
    +---------------------------+-------------------------------------------+
    | vert_vel_desc             | The vertical velocity description         |
    +---------------------------+-------------------------------------------+
    | chan_loc_cd               | The channel location code                 |
    +---------------------------+-------------------------------------------+
    | chan_loc_dist             | The channel location distance             |
    +---------------------------+-------------------------------------------+

    https://help.waterdata.usgs.gov/output-formats#streamflow_measurement_data

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(
        nwis_measurements(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
        )
    )


def nwis_measurements(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    r"""Download from the Measurements database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://nwis.waterdata.usgs.gov/XX/nwis/measurements"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_peak", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_peak_cli(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    r"""Download from the Peak database of the USGS NWIS.

    {filter_descriptions}
    **Results**

    +---------------+--------------------------------------------+
    | Column        | Description                                |
    +===============+============================================+
    | agency_cd     | Agency Code                                |
    +---------------+--------------------------------------------+
    | site_no       | USGS station number                        |
    +---------------+--------------------------------------------+
    | peak_dt       | Date of peak streamflow                    |
    |               |   (format YYYY-MM-DD)                      |
    +---------------+--------------------------------------------+
    | peak_tm       | Time of peak streamflow                    |
    |               |   (24 hour format, 00:00 - 23:59)          |
    +---------------+--------------------------------------------+
    | peak_va       | Annual peak streamflow value in cfs        |
    +---------------+--------------------------------------------+
    | peak_cd       | Peak Discharge-Qualification codes         |
    |               |   (see explanation below)                  |
    +---------------+--------------------------------------------+
    | gage_ht       | Gage height for the associated peak        |
    |               |   streamflow in feet                       |
    +---------------+--------------------------------------------+
    | gage_ht_cd    | Gage height qualification codes            |
    +---------------+--------------------------------------------+
    | year_last_pk  | Peak streamflow reported is the highest    |
    |               |   since this year                          |
    +---------------+--------------------------------------------+
    | ag_dt         | Date of maximum gage-height for water year |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_tm         | Time of maximum gage-height for water year |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_gage_ht    | maximum Gage height for water year in feet |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_gage_ht_cd | maximum Gage height code                   |
    +---------------+--------------------------------------------+

    Peak Streamflow-Qualification Codes(peak_cd):
    +---------+---------------------------------------------------+
    | peak_cd | Description                                          |
    +=========+======================================================+
    | 1       | Discharge is a Maximum Daily Average                 |
    +---------+------------------------------------------------------+
    | 2       | Discharge is an Estimate                             |
    +---------+------------------------------------------------------+
    | 3       | Discharge affected by Dam Failure                    |
    +---------+------------------------------------------------------+
    | 4       | Discharge less than indicated value,                 |
    |         |   which is Minimum Recordable Discharge at this site |
    +---------+------------------------------------------------------+
    | 5       | Discharge affected to unknown degree by              |
    |         |   Regulation or Diversion                            |
    +---------+------------------------------------------------------+
    | 6       | Discharge affected by Regulation or Diversion        |
    +---------+------------------------------------------------------+
    | 7       | Discharge is an Historic Peak                        |
    +---------+------------------------------------------------------+
    | 8       | Discharge actually greater than indicated value      |
    +---------+------------------------------------------------------+
    | 9       | Discharge due to Snowmelt, Hurricane,                |
    |         |   Ice-Jam or Debris Dam breakup                      |
    +---------+------------------------------------------------------+
    | A       | Year of occurrence is unknown or not exact           |
    +---------+------------------------------------------------------+
    | B       | Month or Day of occurrence is unknown or not exact   |
    +---------+------------------------------------------------------+
    | C       | All or part of the record affected by Urbanization,  |
    |         |    Mining, Agricultural changes, Channelization,     |
    |         |    or other                                          |
    +---------+------------------------------------------------------+
    | D       | Base Discharge changed during this year              |
    +---------+------------------------------------------------------+
    | E       | Only Annual Maximum Peak available for this year     |
    +---------+------------------------------------------------------+

    Gage height qualification codes(gage_ht_cd,ag_gage_ht_cd):
    +---------------+------------------------------------------------+
    | gage_ht_cd    | Description                                    |
    | ag_gage_ht_cd |                                                |
    +===============+================================================+
    | 1             | Gage height affected by backwater              |
    +---------------+------------------------------------------------+
    | 2             | Gage height not the maximum for the year       |
    +---------------+------------------------------------------------+
    | 3             | Gage height at different site and(or) datum    |
    +---------------+------------------------------------------------+
    | 4             | Gage height below minimum recordable elevation |
    +---------------+------------------------------------------------+
    | 5             | Gage height is an estimate                     |
    +---------------+------------------------------------------------+
    | 6             | Gage datum changed during this year            |
    +---------------+------------------------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(
        nwis_peak(
            sites=sites,
            stateCd=stateCd,
            huc=huc,
            bBox=bBox,
            countyCd=countyCd,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
        )
    )


def nwis_peak(
    sites=None,
    stateCd=None,
    huc=None,
    bBox=None,
    countyCd=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
):
    r"""Download from the Peak database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://nwis.waterdata.usgs.gov/XX/nwis/peak"
    r = resource(
        url,
        sites=sites,
        stateCd=stateCd,
        huc=huc,
        bBox=bBox,
        countyCd=countyCd,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
    )

    return odo(r, pd.DataFrame)


@mando.command("nwis_stat", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def nwis_stat_cli(
    sites=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    statReportType=None,
    statType=None,
    missingData=None,
    statYearType=None,
):
    r"""Download from the Statistic database of the USGS NWIS.

    {filter_descriptions}
    **Returns**

    This returns a pandas DataFrame if using the Python API, or a text table to
    standard out if using the command line interface.  Both will have columns
    named according to the following table.

    +--------------+----------------+-----------------------------------------+
    | Column Name  | statReportType | Description                             |
    +==============+================+=========================================+
    | agency_cd    | all            | Agency code                             |
    +--------------+----------------+-----------------------------------------+
    | site_no      | all            | Site identification number              |
    +--------------+----------------+-----------------------------------------+
    | parameter_cd | all            | Parameter code                          |
    +--------------+----------------+-----------------------------------------+
    | station_nm   | all            | Site name                               |
    +--------------+----------------+-----------------------------------------+
    | loc_web_ds   | all            | Additional measurement description      |
    +--------------+----------------+-----------------------------------------+
    | year_nu      | monthly        | The year for which the statistics       |
    |              | annual         | apply.                                  |
    +--------------+----------------+-----------------------------------------+
    | month_nu     | monthly        | The month for which the statistics      |
    |              | daily          | apply.                                  |
    +--------------+----------------+-----------------------------------------+
    | day_nu       | daily          | The day for which the statistics apply. |
    +--------------+----------------+-----------------------------------------+
    | begin_yr     | daily          | First water year of data of daily mean  |
    |              |                | values for this day.                    |
    +--------------+----------------+-----------------------------------------+
    | end_yr       | daily          | Last water year of data of daily mean   |
    |              |                | values for this day.                    |
    +--------------+----------------+-----------------------------------------+
    | count_nu     | all            | Number of values used in the            |
    |              |                | calculation.                            |
    +--------------+----------------+-----------------------------------------+
    | max_va_yr    | daily          | Water year in which the maximum value   |
    |              |                | occurred.                               |
    +--------------+----------------+-----------------------------------------+
    | max_va       | daily          | Maximum of daily mean values for        |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | min_va_yr    | daily          | Water year in which the minimum value   |
    |              |                | occurred.                               |
    +--------------+----------------+-----------------------------------------+
    | min_va       | daily          | Minimum of daily mean values for        |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | mean_va      | all            | Mean of daily mean values for this day. |
    +--------------+----------------+-----------------------------------------+
    | p05_va       | daily          | 05 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p10_va       | daily          | 10 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p20_va       | daily          | 20 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p25_va       | daily          | 25 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p50_va       | daily          | 50 percentile (median) of daily         |
    |              |                | mean values for this day.               |
    +--------------+----------------+-----------------------------------------+
    | p75_va       | daily          | 75 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p80_va       | daily          | 80 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p90_va       | daily          | 90 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p95_va       | daily          | 95 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+

    Parameters
    ----------
    {sites}
    {agencyCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {drainAreaMin}
    {drainAreaMax}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}
    {statReportType}
    {statType}
    {missingData}
    {statYearType}"""
    tsutils._printiso(
        nwis_stat(
            sites=sites,
            parameterCd=parameterCd,
            siteType=siteType,
            modifiedSince=modifiedSince,
            agencyCd=agencyCd,
            siteStatus=siteStatus,
            altMin=altMin,
            altMax=altMax,
            drainAreaMin=drainAreaMin,
            drainAreaMax=drainAreaMax,
            aquiferCd=aquiferCd,
            localAquiferCd=localAquiferCd,
            wellDepthMin=wellDepthMin,
            wellDepthMax=wellDepthMax,
            holeDepthMin=holeDepthMin,
            holeDepthMax=holeDepthMax,
            period=period,
            startDT=startDT,
            endDT=endDT,
            statReportType=statReportType,
            statType=statType,
            missingData=missingData,
            statYearType=statYearType,
        )
    )


def nwis_stat(
    sites=None,
    parameterCd=None,
    period=None,
    startDT=None,
    endDT=None,
    siteType=None,
    modifiedSince=None,
    agencyCd=None,
    siteStatus=None,
    altMin=None,
    altMax=None,
    drainAreaMin=None,
    drainAreaMax=None,
    aquiferCd=None,
    localAquiferCd=None,
    wellDepthMin=None,
    wellDepthMax=None,
    holeDepthMin=None,
    holeDepthMax=None,
    statReportType=None,
    statType=None,
    missingData=None,
    statYearType=None,
):
    r"""Download from the Statistic database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder

    url = r"http://waterservices.usgs.gov/nwis/stat/"
    r = resource(
        url,
        sites=sites,
        parameterCd=parameterCd,
        siteType=siteType,
        modifiedSince=modifiedSince,
        agencyCd=agencyCd,
        siteStatus=siteStatus,
        altMin=altMin,
        altMax=altMax,
        drainAreaMin=drainAreaMin,
        drainAreaMax=drainAreaMax,
        aquiferCd=aquiferCd,
        localAquiferCd=localAquiferCd,
        wellDepthMin=wellDepthMin,
        wellDepthMax=wellDepthMax,
        holeDepthMin=holeDepthMin,
        holeDepthMax=holeDepthMax,
        period=period,
        startDT=startDT,
        endDT=endDT,
        statReportType=statReportType,
        statType=statType,
        missingData=missingData,
        statYearType=statYearType,
    )

    return odo(r, pd.DataFrame)


@mando.command("epa_wqp", formatter_class=HelpFormatter, doctype="numpy")
@tsutils.doc(nwis_docstrings)
def epa_wqp_cli(
    bBox=None,
    lat=None,
    lon=None,
    within=None,
    countrycode=None,
    statecode=None,
    countycode=None,
    siteType=None,
    organization=None,
    siteid=None,
    huc=None,
    sampleMedia=None,
    characteristicType=None,
    characteristicName=None,
    pCode=None,
    activityId=None,
    startDateLo=None,
    startDateHi=None,
):
    r"""Download from the EPA Water Quality Portal.

    All of the keywords could be considered as subtractive filters of the
    stations.  Some of the keywords can result in very large number of stations
    being downloaded.

    Parameters
    ----------
    {bBox}
    lat : float
        [optional, default is None]

        Latitude for radial search, expressed in decimal degrees, WGS84

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    lon : float
        [optional, default is None]

        Longitude for radial search, expressed in decimal degrees, WGS84

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    within : float
        [optional, default is None]

        Distance for radial search, expressed in decimal miles

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    countrycode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code. (see domain service for available codes)

        FIPS country codes were established by the National Institute of
        Standards, publication 5-2.

    statecode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code, followed bu a ":", followed by a two-digit FIPS state code. (see
        domain service for available codes)

        FIPS state codes were established by the National Institute of
        Standards, publication 5-2.

    countycode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code, followed by a ":", followed by a two-digit FIPS state code,
        followed by a ":", followed by a three-digit FIPS county code. (see
        domain service for available codes)

        FIPS county codes were established by the National Institute of
        Standards, publication 5-2.

    siteType : str
        [optional, default is None]

        One or more case-sensitive site types, separated by semicolons. (see
        domain service for available site types)

        Restrict retrieval to stations with specified site type (location in
        the hydrologic cycle).  The MonitoringLocationTypeName for individual
        records may provide more detailed information about the type of
        individual stations.

    organization : str
        [optional, default is None]

        For USGS organization IDs, append an upper-case postal-service state
        abbreviation to "USGS-" to identify the USGS office managing the data
        collection station records. However, a few US states are serviced by
        one USGS office.::

            USGS-MA = Massachusetts and Rhode Island
            USGS-MD = Maryland, Delaware, and the District of Columbia
            USGS-PR = Caribbean Islands
            USGS-HI = Pacific Islands

        (see domain service for available organization IDs)

        USGS offices sometimes provide data for stations outside the political
        boundaries associated with the office's organization code. Use the
        statecode or countycode arguments to search for stations located within
        those political boundaries.

    siteid : str
        [optional, default is None]

        Concatenate an agency code, a hyphen ("-"), and a site-identification
        number.

        Each data collection station is assigned a unique
        site-identification number. Other agencies often use different site
        identification numbers for the same stations.

    {huc}

    sampleMedia : str
        [optional, default is None]

        One or more case-sensitive sample media, separated by semicolons. (see
        domain service for available sample media)

        Sample media are broad general classes, and may be subdivided in the
        retrieved data. Examine the data elements ActivityMediaName,
        ActivityMediaSubdivisionName, and ResultSampleFractionText for more
        detailed information.

    characteristicType : str
        [optional, default is None]

        One or more case-sensitive characteristic types (groupings) separated
        by semicolons. (see domain service for available characteristic types)

        These groups will be expanded as part of the ongoing collaboration
        between USGS and USEPA.

    characteristicName : str
        [optional, default is None]

        One or more case-sensitive characteristic names, separated by
        semicolons. (see domain service for available characteristic names)

        Characteristic names identify different types of environmental
        measurements. The names are derived from the USEPA Substance Registry
        System (SRS). USGS uses parameter codes for the same purpose and has
        associated most parameters to SRS names.

    pCode : str
        [optional, default is None]

        One or more five-digit USGS parameter codes, separated by semicolons.
        This is equivalent to "parameterCd" used in other USGS web services.

    activityId : str
        [optional, default is None]

        One or more case-sensitive activity IDs, separated by semicolons.
        Designator that uniquely identifies an activity within an organization.

    startDateLo : str
        [optional, default is None]

        Date of earliest desired data-collection activity.  A very wide range
        of date strings can be used but the closer to ISO 8601 the better.

    startDateHi : str
        [optional, default is None]

        Date of last desired data-collection activity.  A very wide range of
        date strings can be used but the closer to ISO 8601 the better."""
    tsutils._printiso(
        epa_wqp(
            bBox=bBox,
            lat=lat,
            lon=lon,
            within=within,
            countrycode=countrycode,
            statecode=statecode,
            countycode=countycode,
            siteType=siteType,
            organization=organization,
            siteid=siteid,
            huc=huc,
            sampleMedia=sampleMedia,
            characteristicType=characteristicType,
            characteristicName=characteristicName,
            pCode=pCode,
            activityId=activityId,
            startDateLo=startDateLo,
            startDateHi=startDateHi,
        )
    )


def epa_wqp(
    bBox=None,
    lat=None,
    lon=None,
    within=None,
    countrycode=None,
    statecode=None,
    countycode=None,
    siteType=None,
    organization=None,
    siteid=None,
    huc=None,
    sampleMedia=None,
    characteristicType=None,
    characteristicName=None,
    pCode=None,
    activityId=None,
    startDateLo=None,
    startDateHi=None,
):
    from tsgettoolbox.services.epa import wqp as placeholder

    url = r"https://www.waterqualitydata.us/Result/search"
    r = resource(
        url,
        bBox=bBox,
        lat=lat,
        lon=lon,
        within=within,
        countrycode=countrycode,
        statecode=statecode,
        countycode=countycode,
        siteType=siteType,
        organization=organization,
        siteid=siteid,
        huc=huc,
        sampleMedia=sampleMedia,
        characteristicType=characteristicType,
        characteristicName=characteristicName,
        pCode=pCode,
        activityId=activityId,
        startDateLo=startDateLo,
        startDateHi=startDateHi,
    )

    return odo(r, pd.DataFrame)


nwis.__doc__ = nwis_cli.__doc__

nwis_iv.__doc__ = nwis_iv_cli.__doc__
nwis_dv.__doc__ = nwis_dv_cli.__doc__
nwis_site.__doc__ = nwis_site_cli.__doc__
nwis_gwlevels.__doc__ = nwis_gwlevels_cli.__doc__
nwis_measurements.__doc__ = nwis_measurements_cli.__doc__
nwis_peak.__doc__ = nwis_peak_cli.__doc__
nwis_stat.__doc__ = nwis_stat_cli.__doc__
epa_wqp.__doc__ = epa_wqp_cli.__doc__
